#!/usr/local/bin/env python
from __future__ import division
import sys
import warnings
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd.tools import colors, analyze
from IPython.display import display, Latex
from scipy.stats.mstats import mquantiles
from itertools import product

warnings.filterwarnings("ignore")
warnings.simplefilter('ignore', np.RankWarning)
warnings.filterwarnings("ignore", module="matplotlib")

cdict = colors.get_cpals('all')
rpal = cdict['rpal']
bpal = cdict['bpal']
gpal = cdict['gpal']
ppal = cdict['ppal']
heat = cdict['heat']
cool = cdict['cool']
slate = cdict['slate']


def plot_model_fits(y, yhat, ssd=None, ssderr=None, quantiles=np.arange(.1, 1.,.1), err=None, clrs=None, save=False, bw='scott', savestr=None, same_axis=True, lbls=None, cumulative=True, suppressLegend=False, hist=False, kde=True,  shade=True, norm_hist=True, data=None, simData=None, plot_error_rts=True, figure=None):
    """ main plotting function for displaying model fit predictions over data
    """
    sns.set(style='white', font_scale=1.3)
    if np.ndim(y)==1:
        y = y.reshape(1, -1)
    nlevels = y.shape[0]
    yhat = yhat.reshape(nlevels, -1)
    nquant = quantiles.size
    nssd = y[0, 1:-nquant*2].size
    ncols = 2
    figwidth = 10
    if plot_error_rts:
        ncols = 3
        figwidth = 13

    if figure is not None:
        f = figure
        axes = np.asarray(f.axes).reshape(1, ncols)
    elif same_axis:
        f, axes = plt.subplots(1, ncols, figsize=(figwidth, 4.3))
        axes = axes.reshape(1, ncols)
    else:
        f, axes = plt.subplots(nlevels, ncols, figsize=(figwidth, 4*nlevels))
        axes = axes.reshape(nlevels, ncols)

    if ssd is None and nssd>1:
        ssd = np.vstack([np.linspace(.2, .5, nssd)]*nlevels)
        ssd = ssd.reshape(nlevels, -1)

    if clrs is None:
        clrs = colors.assorted_list()
    if lbls is None:
        lbls = [None]*nlevels

    plot_acc = plot_stop_fit
    if nssd==1:
        plot_acc = plot_stop_fit_single
        if ssderr is None:
            ssderr = [np.array([0])]*nlevels

    saccErr, quantErr = [None]*2

    for i in range(nlevels):
        ax_ix = i
        if same_axis:
            ax_ix = 0
        if plot_error_rts:
            ax1, ax2, ax3 = axes[ax_ix]
            qAxes = np.array([ax2,ax3])
        else:
            ax1, ax2 = axes[ax_ix]
            qAxes = np.array([ax2])
        sacc, quant = unpack_vector(y[i], nlevels=nlevels, nquant=nquant, nssd=nssd)
        saccHat, quantHat = unpack_vector(yhat[i], nlevels=nlevels, nquant=nquant, nssd=nssd)

        if err is not None:
            err = err.reshape(nlevels, -1)
            saccErr, quantErr = unpack_vector(err[i], nlevels=nlevels, nquant=nquant, nssd=nssd)

        if nssd==1:
            plot_acc(y=sacc, yhat=saccHat, yerr=saccErr, x=ssd[i], xerr=ssderr[i], color=clrs[i], label=lbls[i], ax=ax1)
        else:
            plot_acc(sacc, saccHat, x=ssd[i], err=saccErr, color=clrs[i], label=lbls[i], ax=ax1, suppressLegend=suppressLegend)

        if not cumulative:
            plot_quantiles(data=data[i], clr=clrs[i], axes=qAxes, bw=bw, cumulative=cumulative, hist=True, kde=False, shade=True, norm_hist=norm_hist, same_axis=same_axis, alpha=.4)

            plot_quantiles(data=simData[i], clr=clrs[i], axes=qAxes, bw=bw, cumulative=cumulative, hist=False, kde=True, shade=False, norm_hist=norm_hist, same_axis=same_axis, alpha=1.)

        else:
            qpcHat, qpeHat = quantHat
            qpcHatX, qpcHatCurve = analyze.fit_logistic(qpcHat, quantiles)
            ax2.errorbar(quant[0], quantiles, xerr=quantErr[0], marker='o', color=clrs[i], ms=6.5, linewidth=0, elinewidth=2.)
            ax2.plot(qpcHatX, qpcHatCurve, linewidth=2, color=clrs[i])
            ax2.plot(qpcHat, quantiles, lw=0., marker='o', ms=10, mew=0, alpha=.1, mfc=clrs[i], mec=clrs[i], color=clrs[i])
            ax2.plot(qpcHat, quantiles, lw=0., marker='o', ms=10, mew=1.5, alpha=.8, mfc='none', mec=clrs[i])
            if plot_error_rts:
                qpeHatX, qpeHatCurve = analyze.fit_logistic(qpeHat, quantiles)
                ax3.errorbar(quant[1], quantiles, xerr=quantErr[1], marker='o', color=clrs[i], ms=6.5, linewidth=0, elinewidth=2.)
                ax3.plot(qpeHatX, qpeHatCurve, linewidth=2, color=clrs[i])
                ax3.plot(qpeHat, quantiles, lw=0., marker='o', ms=10, mew=0, alpha=.1, mfc=clrs[i], mec=clrs[i], color=clrs[i])
                ax3.plot(qpeHat, quantiles, lw=0., marker='o', ms=10, mew=1.5, alpha=.8, mfc='none', mec=clrs[i])

        format_rt_axes(qAxes, cdf=cumulative, yhat=yhat, quantiles=quantiles)
        if nssd==1:
            ax1.set_xlim(np.min(ssd)*.99, np.max(ssd)*1.01)
            xt = np.linspace(np.min(ssd)*.99, np.max(ssd)*1.01,4)
            xtl = [np.int(t*1000) for t in xt]
            ax1.set_xticks(xt)
            ax1.set_xticklabels(xtl)
    if save:
        plt.savefig('.'.join([savestr, 'png']), dpi=600)
    # plt.legend()



def plot_stop_fit(y, yhat, x=None, err=None, color=None, label=None, ax=None, alpha=1., m='o', suppressLegend=False):
    """ plotting function for displaying model-predicted
    stop curve (across SSDs) over empirical estimates
    """
    if ax is None:
        f, ax = plt.subplots(1, figsize=(4, 3))
    if x is None:
        x = np.linspace(.100, .500, len(y), dtype='float')
    if label is None:
        label = ['Data', 'Model']
    else:
        label = [label.capitalize(), None]
    if suppressLegend:
        label = [None, None]
    plot_stop_data(y, x=x, err=err, color=color, label=label[0], ax=ax, lw=0)
    plot_stop_curve_predicted(yhat, x=x, color=color, label=label[1], ax=ax, alpha=alpha)
    format_stop_axes(ax, x)
    # plt.legend()
    sns.despine()



def plot_stop_data(y, x=None, err=None, label=None, lw=2, alpha=1, color='k', ax=None, **kwargs):
    """
    PLOT empirical stop curve
    """
    if ax is None:
        f, ax = plt.subplots(figsize=(4,3))
    if x is None:
        x = np.linspace(.250, .50, len(y))
    if err is not None:
        ax.errorbar(x, y, yerr=err, lw=lw, elinewidth=2, color=color, marker='o', ms=6.5, alpha=alpha, label=label)
    else:
        ax.plot(x, y, lw=lw, color=color, marker='o', ms=6.5, alpha=alpha, label=label)
    sns.despine()
    if label is not None:
        ax.legend()
    return ax



def plot_stop_curve_predicted(y, x=None, label=None, alpha=1., color='k', ax=None, get_pse=False, **kwargs):
    """
    PLOT model-estimated stop curve
    """
    if ax is None:
        f, ax = plt.subplots(figsize=(4,3))
    if x is None:
        x = np.linspace(.250, .50, len(y))
    # xsim, ysim = analyze.fit_sigmoid(x, y)
    xsim, ysim = analyze.fit_logistic(x[::-1], y)
    ax.plot(x, y, lw=0., marker='o', ms=10, mew=1.5, alpha=.1, mfc=color, mec=color)
    ax.plot(x, y, lw=0., marker='o', ms=10, mew=1.5, alpha=.8, mfc='none', mec=color)
    ax.plot(xsim[::-1], ysim, lw=2., color=color, alpha=1, linestyle='-')
    ax.plot(x[0], y[0], lw=2., marker='o', ms=10, mew=1.5, alpha=.8, mfc='none', mec=color, label=label, color=color)
    if get_pse:
        PSEix = (np.abs(ysim - .5)).argmin()
        return xsim[PSE]
    sns.despine()
    return ax



def plot_quantiles(data, axes=None, clr=None, lbl=None, bw='scott', alpha=1., cumulative=True, hist=False, kde=True, shade=True, norm_hist=True, same_axis=False, ms=6.5, mcolor=None, quantiles=np.arange(.1, 1.,.1)):
    goData=data[data.response==1]
    cor = goData[goData['acc']==1]
    err = goData[goData['acc']==0]
    clrCor, clrErr = clr, clr
    if clr is None:
        clrCor, clrErr = "#27ae60",'#de143d'
    if axes is None:
        if same_axis:
            f, axCor = plt.subplots(1, figsize=(5, 4))
            axErr = axCor
        else:
            f, (axCor, axErr) = plt.subplots(1, 2, figsize=(10, 4))
    else:
        axCor, axErr = axes
    if cumulative:
        plot_rt_cdf(cor, ax=axCor, clr=clrCor, lbl=lbl, alpha=alpha, ms=ms, mcolor=mcolor, quantiles=quantiles, bw=bw)
        plot_rt_cdf(err, ax=axErr, clr=clrErr, lbl=lbl, alpha=alpha, ms=ms, mcolor=mcolor, quantiles=quantiles, bw=bw)
    else:
        plot_rt_pdf(cor, ax=axCor, clr=clrCor, lbl=lbl, hist=hist, kde=kde, shade=shade, norm_hist=norm_hist, bw=bw, alpha=alpha)
        plot_rt_pdf(err, ax=axErr, clr=clrErr, hist=hist, kde=kde, shade=shade, norm_hist=norm_hist, bw=bw, alpha=alpha)
        format_rtPDF_axes([axCor, axErr], cor, err)



def plot_rt_cdf(data, ax=None, clr='blue', mcolor=None, lbl=None, alpha=1., ms=6.5, quantiles=np.arange(.1, 1.,.1), lw=2., mew=1., bw='scott'):
    data = data.copy()
    if 'idx' not in data.columns:
        data['idx'] = 1
    if ax is None:
        f, ax = plt.subplots(1, figsize=(4, 3))
    if mcolor is None:
        mcolor = clr
        lw = 0.
        mew = 0
    idxQuant = pd.DataFrame([mquantiles(idxdf.rt.values, quantiles) for idx, idxdf in data.groupby('idx')])
    quant, quantErr = idxQuant.mean(axis=0), idxQuant.sem(axis=0)*1.96
    ax.errorbar(quant, quantiles, xerr=quantErr, color=clr, linewidth=0, elinewidth=1.5, marker='o', ms=ms, mfc=clr, mew=0., label=lbl, mec=clr, alpha=alpha)
    ax.errorbar(quant, quantiles, xerr=quantErr, color=clr, linewidth=0, elinewidth=1.5, marker='o', ms=ms, mfc='none', mew=mew, label=lbl, mec=clr, alpha=.7)
    sns.kdeplot(data.rt.values, cumulative=True, color=clr, linewidth=lw, ax=ax, alpha=1., bw=bw)
    ax.set_xlim(quant.min()*.93, quant.max()*1.05)
    ax.set_ylim(-0.02, 1.02)



def plot_rt_pdf(data, ax=None, clr='blue', lbl=None, alpha=.5, hist=False, kde=True, shade=True, norm_hist=True, bw=.01, bins=18):
    if ax is None:
        f, ax = plt.subplots(1, figsize=(4, 3))
    sns.distplot(data.rt.values, color=clr, label=lbl, kde=True, hist=False, kde_kws={'shade':shade, 'alpha':alpha, 'bw':bw}, norm_hist=norm_hist, hist_kws={'alpha':alpha}, ax=ax, bins=bins)


def plot_stop_fit_single(y, yhat, yerr=None, x=0, xerr=None, ax=None, linestyles=None, color=None, label=None, **kwargs):

    if label is None:
        label = ['Data', 'Model']
    else:
        label = [label, None]
    y = y[1]; yerr=yerr[1]; yhat=yhat[1]
    yline = np.linspace(y-yerr, y+yerr, 4)
    xline = np.ones(yline.size)*x

    ax.plot(x, y, marker='o', ms=6.5, color=color, label=label[0], linewidth=0)
    ax.fill_betweenx(yline, xline-xerr, xline+xerr, alpha=.25, color=color)
    ax.plot(x, yhat, marker='o', ms=10, mew=1.5, alpha=.8, mfc='none', mec=color, label=label[1], lw=0)
    ax.plot(x, yhat, marker='o', ms=10, mew=0, alpha=.1, mfc=color, mec=color)

    # ax.legend(loc=0)
    plt.tight_layout()
    ax.set_ylim(0, 1)



def unpack_vector(vector, nlevels=1, nquant=9, nssd=5):
    if nssd>1:
        # get accuracy at each SSD
        sacc = vector[1:1+nssd]
    else:
        # get go, stop accuracy
        sacc = vector[:2]
    quant = vector[-nquant*2:]
    qc = quant[:nquant]
    qe = quant[-nquant:]
    return sacc, np.vstack([qc, qe])



def format_stop_axes(ax, xticks, xlim=None):
    if xlim is None:
        xlim=(xticks[0]*.935, xticks[-1]*1.05)
    xtls = [np.int(np.round(xx,3)*1000) for xx in xticks]
    yticks = np.array([0., .2, .4, .6, .8,  1.])
    plt.setp(ax, xticks=xticks, xlim=xlim, yticks=yticks, ylim=(-.01, 1.05))
    ax.set_ylabel('Stop Accuracy')
    ax.set_xticklabels(xtls)
    ax.set_yticklabels(yticks)
    # ax.legend(loc=0)
    if ax.is_last_row():
        ax.set_xlabel('SSD (ms)')
    plt.tight_layout()
    sns.despine()
    return ax



def format_rtPDF_axes(axes, cor, err):
    df = pd.concat([cor, err])
    df = df.reset_index(drop=True)
    axCor, axErr = axes
    axCor.set_ylabel('P(RT)')
    axErr.set_ylabel('')
    axCor.set_yticklabels([])
    axErr.set_yticklabels([])
    x = [df.rt.min()*.95, df.rt.median(), df.rt.max()*1.05]
    x = [np.asscalar(np.round(xi,2)) for xi in x]
    axCor.set_xticks(x)
    axErr.set_xticks(x)
    axCor.set_xticklabels([str(xi*1000)[:3] for xi in x])
    axErr.set_xticklabels([str(xi*1000)[:3] for xi in x])
    axCor.set_xlim(x[0], x[-1])
    axErr.set_xlim(x[0], x[-1])
    axCor.set_xlabel('Correct RT (ms)')
    axErr.set_xlabel('Error RT (ms)')



def format_rt_axes(axes, cdf=True, yhat=None, quantiles=np.arange(.1, 1.,.1)):
    if yhat is not None:
        rtqList = [unpack_vector(yh, nquant=quantiles.size)[1:] for yh in yhat]
        corquant = np.hstack(np.vstack(rtqList)[:, 0, :])
        errquant = np.hstack(np.vstack(rtqList)[:, 1, :])
        xxticks = [np.linspace(np.nanmin(rtq)*.93, np.nanmax(rtq)*1.06, 5) for rtq in [corquant, errquant]]
        xxticks = [np.array([np.round(xt, 2) for xt in xxt]) for xxt in xxticks]
        xxtls = [np.array([int(xtl*1000) for xtl in xxt]) for xxt in xxticks]
        xxlim = [(xxt[0], xxt[-1]) for xxt in xxticks]
    q_axes = []
    for ax in axes.flatten():
        if not ax.is_first_col():
            q_axes.append(ax)
    if cdf:
        yylim=(-.02, 1.05); yyticks=np.arange(0, 1.2, .2)
        ylabel = 'Cumulative P(Go)'
    else:
        ydata = np.hstack([np.hstack([l.get_ydata() for l in ax.lines]) for ax in axes])
        yyticks = np.array([0, np.nanmax(ydata)*1.05])
        yyticks = np.array([int(np.round(yt)) for yt in yyticks])
        yylim=(-.02, yyticks[-1]);
        ylabel = 'Probability Mass'
    xlabels = ['Correct RT (ms)', 'Error RT (ms)']
    for i, ax in enumerate(axes):
        plt.setp(ax, yticks=yyticks, ylim=yylim, xlabel=xlabels[i], xlim=xxlim[i], xticks=xxticks[i], xticklabels=xxtls[i])
        if ax.is_last_row():
            plt.setp(ax, xlabel=xlabels[i])
        else:
            plt.setp(ax, xlabel='')
    if not cdf:
        axes[0].set_yticklabels(yyticks)
        if len(axes)>1:
            axes[1].set_yticklabels([], fontsize=.05)
    else:
        axes[0].set_yticklabels([], fontsize=.05)
        if len(axes)>1:
            axes[1].set_yticklabels([], fontsize=.05)
    axes[0].set_ylabel(ylabel, fontsize=17)
    plt.tight_layout()
    sns.despine()
    return axes



def compare_nested_models(fitdf, model_ids, yerr=None, plot_stats=True, verbose=False):
    gof = {}
    fitdf = fitdf[fitdf.pvary.isin(model_ids)]
    # print GOF stats for both models
    for i, p in enumerate(model_ids):
        name = parameter_name(p)
        gof[p] = fitdf.loc[i, ['AIC', 'BIC']]
    # Which model provides a better fit to the data?
    aicwinner = model_ids[np.argmin([gof[mid][0] for mid in model_ids])]
    bicwinner = model_ids[np.argmin([gof[mid][1] for mid in model_ids])]
    if verbose:
        print('AIC likes {} model'.format(aicwinner))
        print('BIC likes {} model'.format(bicwinner))
    plot_model_gof(gof, aicwinner, model_ids, yerr=yerr)
    return gof



def plot_model_gof(gof_dict, aicwinner, pvary=None, yerr=None):
    if pvary is None:
        pvary = np.sort(list(gof_dict))
    nmodels = len(pvary)
    if yerr is None:
        yerr = [np.zeros(2) for i in range(nmodels)]
    f, ax = plt.subplots(1, figsize=(10,7))
    # ax.invert_yaxis()
    x = np.arange(1, nmodels*2, 2)
    clrs = [colors.param_color_map(p) for p in pvary]
    lbls = {p: parameter_name(p,True) for p in pvary}
    for i, p in enumerate(pvary):
        yaic, ybic = gof_dict[p][['AIC', 'BIC']]
        aic_err, bic_err = yerr[i]
        lbl = lbls[p]
        if p==aicwinner:
            lbl+='*'
        ax.bar(x[i]-.32, yaic, yerr=aic_err, color=clrs[i], ecolor="#34495e", error_kw={'linewidth':3, 'alpha':1}, alpha=.8, width=.64, align='center', edgecolor=clrs[i], label=lbl)
        ax.bar(x[i]+.32, ybic, yerr=bic_err, color=clrs[i], ecolor="#34495e", error_kw={'linewidth':3, 'alpha':1}, alpha=.65, width=.64, align='center', edgecolor=clrs[i])
    vals = np.hstack(gof_dict.values()).astype(float)
    yylim = (vals.max()*.97, vals.min()*1.07)
    plt.setp(ax, xticks=x, ylim=yylim, xlim=(0, x[-1]+1), ylabel='IC')
    ax.set_xticklabels(['AIC|BIC']*nmodels, fontsize=14)
    sns.despine(bottom=True, top=False)
    ax.legend(loc=0, fontsize=14)
    ax.invert_yaxis()



def plot_param_distributions(p=['a', 'sso', 'ssv', 'tr', 'v', 'xb', 'z'], n=2000, method='random'):
    from radd import theta
    pkeys = np.sort(list(p))
    nparams = pkeys.size
    p_dists = theta.random_inits(pkeys=pkeys, ninits=n, method=method)
    clrs = colors.param_color_map('all')
    lbls = {pk: parameter_name(pk,True) for pk in pkeys}
    ncols = np.ceil(nparams/2.).astype(int)
    fig, axes = plt.subplots(2, ncols, figsize=(10, 4.5))
    axes = axes.flatten()
    for i, pk in enumerate(pkeys):
        sns.distplot(p_dists[pk], kde=False, ax=axes[i], norm_hist=True, color=clrs[pk], label=lbls[pk], hist_kws={'alpha':.8})
        sns.kdeplot(p_dists[pk], ax=axes[i], color=clrs[pk], shade=True, linewidth=0, alpha=.5, bw='scott')
    for ax in axes:
        ax.legend(loc=1, fontsize=15)
    plt.tight_layout()
    sns.despine()



def compare_param_estimates(p1, p2, depends_on=None):
    popt1 = deepcopy(p1)
    popt2 = deepcopy(p2)
    if depends_on is not None:
        _ = [popt1.pop(param) for param in list(depends_on)]
    pnames = np.sort(list(popt1))
    x = np.arange(pnames.size)
    clrs = sns.color_palette(palette='muted', n_colors=x.size)
    i = 0
    f, ax = plt.subplots(1, figsize=(6,5))
    for param in pnames:
        plt.plot(x[i], abs(popt1[param]), marker='o', markersize=10, color=clrs[i])
        plt.plot(x[i], abs(popt2[param]), marker='s', markersize=8, color=clrs[i])
        i+=1
    _ = plt.setp(plt.gca(), xlim=(x[0]-.5, x[-1]+.5), xticks=x, xticklabels=pnames)



def get_plot_labels(clmap=None):
    nconds = len(list(clmap))
    clevels = [list(levels) for levels in listvalues(clmap)]
    if nconds>1:
        level_data = list(product(*clevels))
        clevels = [' '.join([str(lvl).capitalize() for lvl in lvls]) for lvls in level_data]
    else:
        clevels = np.hstack(clevels).tolist()
        clevels = [lvl.capitalize() for lvl in clevels]
    return clevels


def parameter_name(param, tex=False):
    ix = 0
    if tex:
        ix = 1
    param_name = {'v':['Drift', '$v_{E}$'],
        'ssv': ['Brake Drift', '$v_{B}$'],
        'a': ['Threshold', '$a$'],
        'tr': ['Onset', '$tr$'],
        'xb': ['Dynamic Gain', '$\gamma$'],
        'sso': ['Brake Onset', '$so_{B}$'],
        'z': ['Execution Baseline', '$z_{E}$'],
        #'v_ssv': ['Drift Ratio', '$v_{E},v_{B}$'],
        'aG': ['Alpha+', '$\\alpha^+$'],
        'aErr': ['Alpha-', '$\\alpha^-$'],
        'A': ['Alpha', '$\\beta$'],
        'B': ['Beta', '$\\alpha$'],
        'R': ['Rho', '$\\rho$'],
        'flat': ['Flat', 'Flat'],
        'all': ['Flat', 'Flat']}
    if '_' in param:# and param!='v_ssv':
        param = param.split('_')
    if isinstance(param, list):
        return ','.join([param_name[p][ix] for p in param])
    return param_name[param][ix]



#!/usr/local/bin/env python
from __future__ import division
import os
from future.utils import listvalues
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from numpy.random import randint
import warnings
warnings.simplefilter('ignore', np.RankWarning)
warnings.filterwarnings("ignore", module="matplotlib")

def get_cpals(name='all', aslist=False, random=False):
    rpal = lambda nc: sns.blend_palette(['#e88379', '#de143d'], n_colors=nc)
    bpal = lambda nc: sns.blend_palette(['#81aedb', '#3572C6'], n_colors=nc)
    gpal = lambda nc: sns.blend_palette(['#65b88f', '#27ae60'], n_colors=nc)
    ppal = lambda nc: sns.blend_palette(['#9B59B6', "#663399"], n_colors=nc)
    heat = lambda nc: sns.blend_palette(['#f39c12', '#e5344a'], n_colors=nc)
    cool = lambda nc: sns.blend_palette(["#4168B7", "#27ae60"], n_colors=nc)
    slate = lambda nc: sns.blend_palette(['#95A5A6', "#6C7A89"], n_colors=nc)
    wet = lambda nc: sns.blend_palette(['#34495e', "#99A4AE"], n_colors=nc)
    fire = lambda nc: sns.blend_palette(['#e5344a', "#f39c12"], n_colors=nc)
    bupu = lambda nc: sns.blend_palette(['#8E44AD', "#3498db"], n_colors=nc)
    color_dict = {'bpal': bpal, 'gpal': gpal, 'rpal': rpal, 'ppal': ppal, 'heat': heat, 'cool': cool, 'slate': slate, 'wet': wet, 'fire':fire, 'bupu': bupu}
    if random:
        pals = listvalues(color_dict)
        i = randint(0, len(pals), 1)
        return pals[i]
    if name=='all':
        if aslist:
            return listvalues(color_dict)
        return color_dict
    else:
        return color_dict[name]

def style_params():
    colors = ["#1abc9c", "#16a085", "#2ecc71", "#27ae60", '#009e07', '#94c273', '#83a83b',
              "#3498db", "#2980b9", '#3572C6', "#4168B7", "#3A539B",
              '#8172b2', "#9B59B6", "#8E44AD", "#674172", "#663399",
              "#95A5A6", '#6C7A89', "#34495e", "#2c3e50",
              '#E26A6A', "#e16256", "#e74c3c", "#e5344a", '#de143d', "#c0392b",
              "#f1c40f", "#f39c12", "#e67e22", "#ff914d", "#ff711a"]
    return {'colors': colors, 'greens': colors[:4], 'blues': colors[4:8], 'purples': colors[8:12], 'grays': colors[12:17], 'reds': colors[17:22], 'yellows': colors[22:]}

def param_color_map(param='all'):
    param_color_map = {'a': "#375ee1", 'tr': "#f19b2c", 'v': "#27ae60", 'xb': "#16a085", 'ssv': "#e5344a", 'ssv_v': "#3498db", 'sso': "#e941cd", 'z': '#ff711a', 'all': '#6C7A89', 'flat': '#6C7A89', 'C': '#009e07', 'B': '#de143d', 'Beta': "#ff711a", 'v_ssv': "#9B59B6"}
    if param=='all':
        return param_color_map
    if param in list(param_color_map):
        return param_color_map[param]
    elif '_' in param:
        params = param.split('_')
        blended = [param_color_map[p] for p in params]
        return sns.blend_palette(blended, n_colors=6)[3]
    elif param not in list(param_color_map):
        clrs = assorted_list()
        ix = np.random.randint(0, len(clrs))
        return clrs[ix]

def assorted_list():
    return ['#3572C6',  '#c44e52', '#8172b2', '#83a83b', "#3498db", "#e5344a", '#94c273', '#6C7A89', "#8E44AD", "#16a085", "#f39c12", "#4168B7", '#34495e', "#27ae60", "#e74c3c", "#ff711a", "#ff914d"]

def random_colors(n):
    colornames = list(sns.crayons)
    r_ints = randint(0, high=len(colornames), size=n)
    cnames = [colornames[i] for i in r_ints]
    return [sns.crayons[name] for name in cnames]

def get_cmaps():
    block = ['Blues', 'BuGn', 'BuPu', 'GnBu', 'Greens', 'Greys', 'Oranges', 'OrRd', 'PuBu',
             'PuBuGn', 'PuRd', 'Purples', 'RdPu', 'Reds', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd'],
    seq = ['afmhot', 'autumn', 'bone', 'cool', 'copper',
           'gist_heat', 'gray', 'hot', 'pink', 'summer', 'winter']
    div = ['BrBG', 'bwr', 'coolwarm', 'PiYG', 'PRGn', 'PuOr',
           'RdBu', 'RdGy', 'RdYlBu', 'RdYlGn', 'Spectral', 'seismic']
    qual = ['Accent', 'Dark2', 'Paired', 'Pastel1',
            'Pastel2', 'Set1', 'Set2', 'Set3']

    misc = ['gist_earth', 'terrain', 'ocean', 'gist_stern', 'brg', 'CMRmap', 'cubehelix', 'gnuplot',
            'gnuplot2', 'gist_ncar', 'nipy_spectral', 'jet', 'rainbow', 'gist_rainbow', 'hsv', 'flag', 'prism']
    cmaps = {'block': block, 'div': div,
             'seq': seq, 'qual': qual, 'misc': misc}
    return cmaps



#!/usr/bin/env python
import os
from copy import deepcopy
from future.utils import listvalues
import numpy as np
from numpy.random import randint
from lmfit import fit_report
from time import strftime


def logger(param_report, savepath=None, finfo={}, popt={}, fitparams={}, kind='xdpm', fit_on='average', array_names = ['y', 'wts', 'yhat']):
    """ logs information by opening and appending to an existing log file
    (named according to parameter dependencies) or creating a new log.
    """
    newSection = '-' * 3 + '\n'
    newHeader = lambda x: x + '\n' + '-' * len(x) + '\n'

    if savepath is None:
        savepath = os.path.expanduser('~')
        if 'Dropbox' in os.listdir(savepath):
            savepath = os.path.join(savepath, 'Dropbox')

    # functions for writing numpy arrays to strings (ex. "y = np.array([1,2,3])"")
    name_equals = lambda name, strvector: '{0} = np.array([{1}])'.format(name, strvector)
    stringify = lambda x: name_equals(x[0], ', '.join('{:f}'.format(n) for n in x[1]))
    # brief-ify fitparams reference
    fp = deepcopy(fitparams)
    # list flattened y, wts, and yhat arrays
    arrays = [fp[k].flatten() for k in array_names]
    # write arays to strings for easy logging
    names_arrays = zip(array_names, arrays)
    y_str, wts_str, yhat_str = [stringify(narr) for narr in names_arrays]
    fit_on = '  |  '.join(fp['model_id'].split('_'))

    # limit popt to 6 decimal points
    for k,v in popt.items():
        if isinstance(v, np.ndarray):
            v = np.array([float('{:.6f}'.format(vi)) for vi in v])
        else:
            v = float('{:.6f}'.format(v))
        popt[k] = v
    #popt = {k: float('{:.6f}'.format(v)) for k,v in popt.items()}
    # remove heading from param_report
    param_report = param_report.split(']]')[1]

    if fp['nlevels']==1:
        dep_id = "flat model (no conditional parameters)"
        fname = './' + kind + '_flat.md'
    else:
        depends_on = fp['depends_on']
        dep_id = "\n"
        for p, conds in depends_on.items():
            if not isinstance(conds, list):
                dep_id += "{0} depends on {1}\n".format(p, conds)
            else:
                dep_id += "{0} depends on {1}\n".format(p, ', '.join(np.hstack(conds)))
        dparams = '_'.join(np.hstack(list(depends_on)))
        fname = '_'.join(['./' + kind, dparams + '.md'])

    fname = os.path.abspath(os.path.join(savepath, fname))

    with open(fname, 'a') as f:
        f.write('\n\n')
        f.write(newHeader('MODEL INFO:'))
        f.write(' '.join(['TIMESTAMP:', strftime('%m/%d/%y %I:%M%p'), '\n']))
        f.write(str(fit_on) + '\n')
        f.write(str(dep_id) + '\n\n')
        f.write(newHeader('DATA, YHAT & WEIGHTS:'))
        f.write('\n')
        f.write(wts_str + '\n\n')
        f.write(yhat_str + '\n\n')
        f.write(y_str + '\n')
        f.write('\n\n')
        f.write(newHeader("FIT REPORT:"))
        f.write(param_report + '\n')
        f.write('\n')
        f.write(' = '.join(['popt', repr(popt)]) + '\n')
        f.write('\n\n')
        f.write('ndata: {}\n'.format(finfo['ndata']))
        f.write('df: {}\n'.format(finfo['df']))
        f.write('nfev: {}\n'.format(finfo['nfev']))
        f.write('niter: {}\n'.format(finfo['niter']))
        f.write('chi: {:.8f}\n'.format(finfo['chi']))
        f.write('rchi: {:.8f}\n'.format(finfo['rchi']))
        f.write('logp: {:.4f}\n'.format(finfo['logp']))
        f.write('AIC: {:.4f}\n'.format(finfo['AIC']))
        f.write('BIC: {:.4f}\n'.format(finfo['BIC']))
        f.write('converged: %s' % finfo['cnvrg'] + '\n\n')
        f.write(newSection + '\n\n')


def saygo(depends_on={}, cond_map=None, kind='xdpm', fit_on='subjects'):
    """ generate confirmation message that model is prepared and ready to fit.
    repeats structure and important fit details for user to confirm """
    depkeys = describe_model(depends_on)
    if 'x' in kind:
        bias = '(w/ dynamic bias)'
    else:
        bias = ""
    dep = listvalues(depends_on)
    # flatten list of all cond levels
    lbls = ', '.join(sum(listvalues(depends_on), []))
    msg = get_nonsense()
    strings = (kind, bias, fit_on, pdep, dep, lbls, msg)
    print("""
      Model is prepared to fit %s model %s to %s data,
      allowing %s to vary across levels of %s (%s)
      %s \n""" % strings)
    return True


def describe_model(depends_on=None):
    """ get names of any conditional parameters included in the model
    """
    pkeys = list(depends_on)
    deplist = []
    if 'a' in pkeys:
        deplist.append('boundary')
    if 'tr' in pkeys:
        deplist.append('onset')
    if 'v' in pkeys:
        deplist.append('drift')
    if 'xb' in pkeys:
        deplist.append('xbias')
    if len(pkeys) > 1:
        depkeys = ' and '.join(deplist)
    else:
        depkeys = deplist[0]
    return depkeys

def global_logger(logs):
    arr_str = lambda x: ', '.join(str(elem)[:6] for elem in x)
    str_str = lambda x: ', '.join(str(elem) for elem in x)
    single_str = lambda x: ', '.join(str(elem) for elem in x)
    # create a dictionary string that can be copied and pasted into cell
    #pdict=[':'.join(["'"+str(k)+"'",str(v)]) for k,v in logs['popt'].items()]
    popt = ''.join(['popt =', repr(logs['popt'])])
    fmin_str = 'fmin = %s' % str(logs['fmin'])[:6]
    yhat_str = 'yhat = array([' + arr_str(logs['yhat']) + '])'
    y_str = 'y = array([' + arr_str(logs['y']) + '])'
    cost_str = 'cost = %s' % str(logs['cost'])[:6]
    with open('global_min_report.txt', 'a') as f:
        f.write('==' * 20 + '\n')
        f.write(popt + '\n')
        f.write(fmin_str + '\n')
        f.write(cost_str + '\n')
        f.write(yhat_str + '\n')
        f.write(y_str + '\n')
        f.write('==' * 20 + '\n')

def get_nonsense():
    """ random nonsensical words of encouragement
    """
    msgs = ["Optimize On, Wayne",
            "Optimize On, Garth",
            "See, it's not that simplex...",
            "I wish you a merry fit, and a happy Nature paper",
            "It'll probably work this time",
            "'They dont think it be like it is, but it do' -Oscar Gamble",
            "Check your zoomfile before optimizing. It should be in your computer",
            "It's IN the computer?",
            "What is this... a model for ANTS!?"]
    return msgs[randint(0, len(msgs))]



#!usr/bin/env python
from __future__ import division
from future.utils import listvalues
import pandas as pd
import numpy as np
from copy import deepcopy
from numpy import array
from scipy.stats.distributions import norm, gamma, uniform
from lmfit import Parameters as lmParameters
from pyDOE import lhs


def random_inits(pkeys, ninits=1, kind='dpm', as_list=False, force_normal=False, method='random'):
    """ random parameter values for initiating model across range of
    parameter values.
    ::Arguments::
        pkeys (list/dict):
            list of parameter names to sample, makes list of dict keys
            if pkeys is dictionary type
    ::Returns::
        params (dict):
            dictionary with pkeys as keys with values as 1d arrays ninits long
            of randomly sampled parameter values
    """
    if isinstance(pkeys, dict):
        pkeys = list(pkeys)
    params = {}
    if method=='random':
        for pk in pkeys:
            params[pk] = init_distributions(pk, nrvs=ninits, kind=kind, force_normal=force_normal)
    elif method=='lhs':
        params = latin_hypercube(pkeys, nrvs=ninits, kind=kind)
    if as_list:
        params = np.array([{pk: params[pk][i] for pk in pkeys} for i in range(ninits)])
    return params


def latin_hypercube(pkeys, kind='dpm', nrvs=25, tb=.65, force_normal=False):
    """ sample random parameter sets to explore global minima (called by
    Optimizer method __hop_around__())
    """
    nparams = len(pkeys)
    design = lhs(nparams, samples=nrvs, criterion='center')
    bounds = get_bounds(kind=kind)
    # reverse V_brake boundaries to get max negative val
    bounds['ssv'] = (bounds['ssv'][1], bounds['ssv'][0])
    pmax = np.array([bounds[pk][-1] for pk in pkeys])
    design = pmax * design
    samplesLH = {p: design[:, i] for i, p in enumerate(pkeys)}
    return samplesLH



def init_distributions(pkey, kind='dpm', nrvs=25, tb=.65, force_normal=False):
    """ sample random parameter sets to explore global minima (called by
    Optimizer method __hop_around__())
    """
    loc, scale = get_theta_params(pkey, kind=kind)
    bounds = get_bounds(kind=kind)[pkey]
    lower = np.min(bounds)
    upper = np.max(bounds)
    normal_params = ['a', 'tr', 'v', 'vd', 'ssv', 'sso', 'xb', 'z', 'Beta']
    uniform_params = ['vi', 'C', 'B', 'R', 'si']

    # init and freeze dist shape
    if pkey in normal_params:
        dist = norm(loc, scale)
    # elif pkey in gamma_params:
    #     dist = gamma(1.0, loc, scale)
    elif pkey in uniform_params:
        dist = uniform(loc, scale)
    # generate random variates
    rvinits = dist.rvs(nrvs)
    while rvinits.min() < lower:
        # apply lower limit
        ix = rvinits.argmin()
        rvinits[ix] = dist.rvs()
    while rvinits.max() > upper:
        # apply upper limit
        ix = rvinits.argmax()
        rvinits[ix] = dist.rvs()
    if pkey =='tr':
        rvinits = np.abs(rvinits)
    rvinits[rvinits<lower] = lower
    rvinits[rvinits>upper] = upper
    return rvinits


def get_bounds(kind='dpm', tb=None):
    """ set and return boundaries of parameter optimization space
    ::Arguments::
        kind (str): model type
        tb (float): timeboundary
    ::Returns::
        bounds (dict): {parameter: (upper, lower)}
    """

    bounds = {'a': (.3, .6),
             'si': (.001, .15),
             'sso': (.005, .1),
             'ssv': (-1.2, -.4),
             'tr': (0.1, 0.45),
             'v': (.6, 1.2),
             'xb': (.6, 1.4),
             'z': (0.01, 0.6)}

    boundsRL = {'B': (.1, .4),
                'Beta': (.5, 5.),
                'C': (.01, .1),
                'R': (.0001, .005),
                'vd': (.1, 2.1),
                'vi': (.1, 1.0)}

    if tb is not None:
        bounds['tr'] = (bounds['tr'][0], tb-0.1)
    if 'irace' in kind:
        mu, sigma = theta['ssv']
        theta['ssv'] = (abs(mu), sigma)

    bounds.update(boundsRL)
    return bounds


def get_theta_params(pkey, kind='dpm'):
    """ set and return loc, scale of parameter pkey
    ::Arguments::
        pkey (str): parameter name to get loc and scale for sampling
    ::Returns::
        theta[pkey] (tuple): (loc, scale)
    """
    theta = {'a': (.35, .15),
             'si': (.001, .1),
             'sso': (.05, .035),
             'ssv': (-.5, .3),
             'tr': (.25, .075),
             'v': (.8, .3),
             'xb': (1., .35),
             'z': (.3, .2)}

    thetaRL={'B': (.1, .4),
             'C': (.001, .08),
             'R': (.0005, .008),
             'vd': (.5, .5),
             'vi': (.3, .4)}

    if 'irace' in kind:
        mu, sigma = theta['ssv']
        theta['ssv'] = (abs(mu), sigma)

    theta.update(thetaRL)
    return theta[pkey]



def loadParameters(inits=None, is_flat=False, kind=None, pcmap={}):
    """ Generates and returns an lmfit Parameters object with
    bounded parameters initialized for flat or non flat model fit
    """
    lmParams = lmParameters()
    pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso', 'C', 'B', 'R']
    pfit = list(set(inits.keys()).intersection(pnames))
    bounds = get_bounds(kind=kind)
    for pkey, pclist in pcmap.items():
        if is_flat:
            break  # exit
        pfit.remove(pkey)
        if hasattr(inits[pkey], '__iter__'):
            vals = inits[pkey]
        else:
            vals = inits[pkey] * np.ones(len(pclist))
        mn = bounds[pkey][0]
        mx = bounds[pkey][1]
        for k, v in zip(pclist, vals):
            if isinstance(v, np.ndarray):
                print(v)
                v = np.asscalar(v)
            lmParams.add(k, value=v, vary=True, min=mn, max=mx)
    for pk in pfit:
        inits = scalarize_params(inits, pfit)
        if is_flat:
            mn = bounds[pk][0]
            mx = bounds[pk][1]
            lmParams.add(pk, value=inits[pk], vary=True, min=mn, max=mx)
        elif isinstance(inits[pk], np.ndarray):
            pk_clist = ['_'.join([pk, pcond.split('_')[1]]) for pcond in pclist]
            for pname, pvals in zip(pk_clist, inits[pk]):
                lmParams.add(pname, value=pvals, vary=False)
        else:
            lmParams.add(pk, value=inits[pk], vary=False)
    return lmParams


def loadParameters_RL(inits, pvary, pflat, nlevels=1, kind='xdpm'):
    """ Generates and returns an lmfit Parameters object with
    bounded parameters initialized for flat or non flat model fit
    """
    lmParams = lmParameters()
    bounds = get_bounds(kind=kind)
    if nlevels > 1:
        varyIDs = ['{}_{}'.format(pkey, str(l)) for l in np.arange(nlevels)]
    for pkey in pvary:
        mn = bounds[pkey][0]
        mx = bounds[pkey][1]
        for i in range(nlevels):
            varyID = pkey
            if nlevels>1:
                varyID = '{}_{}'.format(pkey, str(l))
            lmParams.add(varyID, value=inits[pkey], vary=True, min=mn, max=mx)
    for pk in pflat:
        lmParams.add(pk, value=inits[pk], vary=False)
    return lmParams


def get_default_inits(kind='dpm', depends_on={}, learn=False, ssdelay=False, gbase=False):
    """ if user does not provide inits dict when initializing Model instance,
    grab default dictionary of init params reasonably suited for Model kind
    """
    inits = {'a': 0.42, 'v': 1.25, 'tr': 0.2}
    pdep = list(depends_on)

    if 'dpm' in kind:
        inits['ssv'] = -1.3
    elif 'race' in kind:
        inits['ssv'] = 1.3
    if 'x' in kind and 'xb' not in list(inits):
        inits['xb'] = 1.5
    if 'si' in pdep:
        inits['si'] = .01
    if ssdelay:
        inits['sso'] = .08
    if gbase:
        inits['z'] = .1
    if learn:
        inits['C'] = .02
        inits['B'] = .15
        inits['R'] = .0015
    return inits


def clean_popt(params, pcmap):
    params = deepcopy(params)
    pvary = np.sort(list(pcmap))
    for i, p in enumerate(pvary):
        p_lvls = np.sort(pcmap[p])
        params[p] = array([params[p_lvl] for p_lvl in p_lvls])
    for p in list(params):
        if p in pvary.tolist():
            continue
        pval = np.mean(params[p])
        params[p] = pval * np.ones(p_lvls.size)
    return params


def scalarize_params(params, pcmap={}):
    """ scalarize all parameters in params dict """
    params = deepcopy(params)
    pvary = np.sort(list(pcmap)).tolist()
    if isinstance(params, pd.Series):
        params = params.to_dict()
    for p in list(params):
        if p in pvary:
            continue
        if hasattr(params[p], '__iter__'):
            try:
                params[p] = np.asscalar(params[p])
            except ValueError:
                params[p] = np.mean(params[p])
    return params


def pvary_levels_to_array(popt, pcmap={}):
    """ store optimized values of parameter in popt dict as array
        Example:
            from:   {'v_bsl': 1.15, 'v_pnl':1.10, ...}
            to:     {'v': array([1.15, 1.10]), ...}
    """
    for pname, pconds in pcmap.items():
        popt[pname] = np.array([popt[c] for c in pconds])
    return popt


def check_inits(inits={}, depends_on={}, kind='dpm'):
    """ ensure inits dict is appropriate for Model kind
    """
    pdep = list(depends_on)
    if 'race' in kind or 'iact' in kind:
        inits['ssv'] = abs(inits['ssv'])
    elif 'dpm' in kind:
        inits['ssv'] = -abs(inits['ssv'])
    if 'pro' in kind:
        if 'ssv' in inits.keys():
            ssv = inits.pop('ssv')
    if 'x' in kind and 'xb' not in inits.keys():
        inits['xb'] = 1.5
    if 'si' in pdep and 'si' not in inits.keys():
        inits['si'] = .01
    if 'dpm' not in kind and 'z' in inits.keys():
        discard = inits.pop('z')
    if 'x' not in kind and 'xb' in inits:
        inits.pop('xb')
    # make sure inits only contains subsets of these params
    pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso', 'C', 'B', 'R']
    pfit = list(set(list(inits)).intersection(pnames))
    return {pk: inits[pk] for pk in pfit}


def update_params(theta):
    if 't_hi' in theta.keys():
        theta['tr'] = theta['t_lo'] + np.random.uniform() * (theta['t_hi'] - theta['t_lo'])
    if 'z_hi' in theta.keys():
        theta['z'] = theta['z_lo'] + np.random.uniform() * (theta['z_hi'] - theta['z_lo'])
    if 'sv' in theta.keys():
        theta['v'] = theta['sv'] * np.random.randn() + theta['v']
    return theta


def get_intervar_ranges(theta):
    """ theta (dict): parameters dictionary
    """
    if 'st' in theta.keys():
        theta['t_lo'] = theta['tr'] - theta['st'] / 2
        theta['t_hi'] = theta['tr'] + theta['st'] / 2
    if 'sz' in theta.keys():
        theta['z_lo'] = theta['z'] - theta['sz'] / 2
        theta['z_hi'] = theta['z'] + theta['sz'] / 2
    return theta



#!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd import theta
from IPython.display import HTML
from matplotlib import animation

def anim_to_html(anim):
    from tempfile import NamedTemporaryFile
    VIDEO_TAG = """<video controls>
             <source src="data:video/x-m4v;base64,{0}" type="video/mp4">
             Your browser does not support the video tag.
            </video>"""
    if not hasattr(anim, '_encoded_video'):
        with NamedTemporaryFile(suffix='.mp4') as f:
            anim.save(f.name, extra_args=['-vcodec', 'libx264', '-pix_fmt', 'yuv420p'])
            video = open(f.name, "rb").read()
        anim._encoded_video = video.encode("base64")
    return VIDEO_TAG.format(anim._encoded_video)


def render_animation(anim):
    plt.close(anim._fig)
    return HTML(anim_to_html(anim))


def animate_dpm(model):
    """ to render animation within a notebook :
        vis.render_animation(vis.animated_dpm_example(MODEL))
    """
    model.set_fitparams(dt=.001)
    params = deepcopy(model.inits)
    bound=theta.scalarize_params(params)['a']
    # generate reactive model simulations
    x, goTraces, brakeTraces, xi, yi, nframes = gen_re_traces(model)
    f, axes = build_decision_axis(onset=x[0][0], bound=bound)
    # axes line object for "go" process
    goLine = [axes[i].plot([], [], linewidth=1.5)[0] for i, n in enumerate(gtraces)]
    # axes line object for "brake" process
    brakeLine = [axes[i].plot([xi[i]], [yi[i]], linewidth=1.5)[0] for i, n in enumerate(straces)]
    f_args = (x, goTraces, goLine, brakeTraces, brakeLine, params, xi, yi)
    anim=animation.FuncAnimation(f, re_animate_multiax, fargs=f_args, frames=nframes, interval=1, blit=True)
    return anim


def build_decision_axis(onset, bound, ssd=np.array([300, 400]), tb=650):
    # init figure, axes, properties
    f, axes = plt.subplots(len(ssd), 1, figsize=(7, 7), dpi=300)
    #f.subplots_adjust(wspace=.05, top=.1, bottom=.1)
    f.subplots_adjust(hspace=.05, top=.99, bottom=.09)
    w = tb + 40
    h = bound
    start = onset - 80
    for i, ax in enumerate(axes):
        plt.setp(ax, xlim=(start - 1, w + 1), ylim=(0 - (.01 * h), h + (.01 * h)))
        ax.vlines(x=ssd[i], ymin=0, ymax=h, color="#e74c3c", lw=2.5, alpha=.5)
        ax.hlines(y=h, xmin=start, xmax=w, color='k')
        ax.hlines(y=0, xmin=start, xmax=w, color='k')
        ax.vlines(x=tb, ymin=0, ymax=h, color='#2043B0', lw=2.5, linestyle='-', alpha=.5)
        ax.vlines(x=start + 2, ymin=0, ymax=h, color='k')
        ax.text(ssd[i] + 10, h * .88, str(ssd[i]) + 'ms', fontsize=19)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
    sns.despine(top=True, right=True, bottom=True, left=True)
    return f, axes


def gen_re_traces(model):
    params = deepcopy(model.inits)
    dvg, dvs, rt, ssrt = model.sim.simulate_traces(params)
    bound = params['a']
    tr = params['tr']
    goOn = model.sim.gOnset[0]
    ssOn = model.sim.ssOnset.squeeze()
    ssOn = [ssOn[2], ssOn[-1]]
    xinit_ss = [ssOn_i-goOn for ssOn_i in ssOn]
    rtSteps = [int(rt[0, i]*1000)+1 for i in [0, 1]]
    ssrtSteps = [int(ssrt[0, 2, 0]*1000)+1, int(ssrt[0, -1, 0]*1000)+1]
    gtraces = [dvg[0, 0, goOn:rtSteps[0]], dvg[0, 1, goOn:rtSteps[-1]]]
    straces = [dvs[0, 2, 0, ssOn[0]:ssrtSteps[0]], dvs[0, -1, 0, ssOn[1]:ssrtSteps[1]]]
    # gtraces = [gt[gt <= bound] for gt in gtraces]
    ssi, integrated, dvgs, dvss = [], [], [], []
    for i, (g, s) in enumerate(zip(gtraces, straces)):
        ssi.append(g[:xinit_ss[i]])
        ss = s[:ssOn[i]]
        s = np.append(g[:xinit_ss[i]], ss[ss >= 0])
        ixmin = np.min([len(g), len(s)])
        dvgs.append(g[:ixmin])
        dvss.append(s[:ixmin])
    nframes = [len(gt) for gt in dvgs]
    x = [np.arange(goOn, goOn + nf) for nf in nframes]
    return [x, dvgs, dvss, xinit_ss, ssi, np.max(nframes)]


def re_animate_multiax(i, x, gtraces, glines, straces, slines, params, xi, yi):
    gcolor = '#27ae60'
    scolor = '#e74c3c'
    for n in range(len(x)):
        ex, gt, gl, st, sl, ix, iy = [xx[n] for xx in [x, gtraces, glines, straces, slines, xi, yi]]
        try:
            gl.set_data(ex[:i + 1], gt[:i + 1])
            gl.set_color(gcolor)
            sl.set_data(ex[ix:i + 1], st[ix:i + 1])
            sl.set_color(scolor)
        except Exception:
            return sl, gl
        #f.savefig('animation_frames/movie/img' + str(i) +'.png', dpi=300)
    return sl, gl



def gen_ddm_traces(model, ntrials=10):
    params = deepcopy(model.inits)
    dvg, dvs, rt, ssrt = model.sim.simulate_traces(params)
    bound = params['a']
    tr = params['tr']
    goOn = model.sim.gOnset[0]
    trials = range(ntrials)
    rtSteps = [int(rt[0, t]*1000)+1 for t in trials]
    gtraces = [dvg[0, t, goOn:rtSteps[t]] for t in trials]
    nframes = [len(gt) for gt in gtraces]
    x = [np.arange(goOn, goOn + nf) for nf in nframes]

    return [x, gtraces, np.max(nframes)]


def animate_ddm_func(i, x, gtrace, glines, line_i):
    gcolor = '#4168B7'
    # for n in range(len(x)):
    # ex, gt, gl = [xx[n] for xx in [x, gtraces, glines]]
    try:
        line_index = line_i[i]
        gl = glines[line_index]
        xx = x[line_index]
        # gt = gtraces[line_index]
        if x==200:
            gl.set_data(x[i:i + 1], gtraces[i:i + 1])
        else:
            gl.set_data(x[:i + 1], gtraces[:i + 1])
        gl.set_color(gcolor)
    except Exception:
        return gl
        #f.savefig('animation_frames/movie/img' + str(i) +'.png', dpi=300)
    return gl



#!/usr/local/bin/env python
from __future__ import division
from future.utils import listvalues, listitems
from copy import deepcopy
import os
import numpy as np
import pandas as pd
from numpy import array
from scipy.stats.mstats import mquantiles as mq
from lmfit import fit_report
from radd.tools import messages, utils, analyze
from radd import theta, vis
from itertools import product



class RADDCore(object):
    """ Parent class for constructing attributes and methods used by
    of Model objects. Not meant to be used directly

    Contains methods for building dataframes, generating observed data vectors
    that are entered into cost function during fitting as well as calculating
    summary measures and weight matrix for weighting residuals during optimization.
    """
    def __init__(self, data=None, kind='xdpm', inits=None, fit_on='average', depends_on={'all':'flat'}, ssd_method=None, weighted=True, verbose=False, custompath=None, nested_models=None, learn=False, bwfactors=None, ssdelay=False, gbase=False, quantiles=np.arange(.1, 1.,.1), presample=False, ksfit=False):
        self.kind = kind
        self.fit_on = fit_on
        self.ssd_method = ssd_method
        self.weighted = weighted
        self.quantiles = quantiles
        self.learn = learn
        self.ssdelay = ssdelay
        self.gbase = gbase
        self.custompath = custompath
        self.data = data.copy()
        # self.data = analyze.remove_outliers(data, 3.5)
        self.tb = analyze.estimate_timeboundary(self.data)
        self.idx = list(self.data.idx.unique())
        self.nidx = len(self.idx)
        self.bwfactors = bwfactors
        self.inits = inits
        self.finished_sampling = False
        self.track_subjects = False
        self.track_basins = False
        self.pbars = None
        self.is_nested = False
        self.ksfit=ksfit
        self.__prepare_fit__(depends_on, presample=presample)


    def __prepare_fit__(self, depends_on, presample=False):
        """ model setup and initiates dataframes. Automatically run when Model object is initialized
        *pcmap is a dict containing parameter names as keys with values
                corresponding to the names given to that parameter in Parameters object
                (see optmize.Optimizer).
        *Parameters (p[pkey]=pval) that are constant across conditions are broadcast as [pval]*n.
                Conditional parameters are treated as arrays with distinct values [V1, V2...Vn], one for
                each condition.
        pcmap (dict): see bound __format_pcmap__ method
        """

        # from radd.optimize import Optimizer
        # from radd.models import Simulator
        from radd import optimize

        self.set_conditions(depends_on, self.bwfactors)

        if self.inits is None:
            self.__get_default_inits__()

        # pcmap (see docstrings)
        self.__format_pcmap__()
        # create model_id string for naming output
        self.generate_model_id()

        # initialize DataHandler & generate I/O dataframes
        self.__make_dataframes__()

        # set fit parameters with default values
        self.set_fitparams()

        # # set ssd info
        # self.__set_ssd_info__()

        # set basinhopping parameters with default values
        self.set_basinparams()

        # initialize optimizer object for controlling fit routines
        # (updated with fitparams/basinparams whenever params are set)
        self.opt = optimize.Optimizer(fitparams=self.fitparams, basinparams=self.basinparams, inits=self.inits, data=self.data)
        self.sim = self.opt.sim

        if presample:
            # sample init params
            self.sample_theta()


    def __make_dataframes__(self):
        """ wrapper for dfhandler.DataHandler.make_dataframes
        """
        from radd.dfhandler import DataHandler
        # initialize dataframe handler
        self.handler = DataHandler(self)
        # make dataframes
        self.handler.make_dataframes()

        # Group dataframe (nsubjects*nconds*nlevels x ndatapoints)
        self.observedDF = self.handler.observedDF.copy()
        self.observedErr = self.handler.observedErr.copy()
        # list (nsubjects long) of data arrays (nconds*nlevels x ndatapoints) to fit
        self.observed = self.handler.observed
        # list of flattened data arrays (averaged across conditions)
        self.observed_flat = self.handler.observed_flat

        # observed frequency bins dataframe
        # self.freqDF = self.handler.make_freq_df()

        # dataframe with same dim as observeddf for storing model predictions
        self.yhatdf = self.handler.yhatdf
        # dataframe with same dim as observeddf for storing fit info (& popt)
        self.fitdf = self.handler.fitdf
        # dataframe with same dim as observeddf for storing optimized params as matrix
        self.poptdf = self.handler.poptdf
        # dataframe containing cost_function wts (see dfhandler docs)
        self.wtsDF = self.handler.wtsDF

        try:
            # dataframe containing ssd's per idx (see dfhandler docs)
            self.ssdDF = self.handler.ssdDF
        except Exception:
            pass

        # list of arrays containing conditional costfx weights
        self.cond_wts = self.handler.cond_wts
        # list of arrays containing flat costfx weights
        self.flat_wts = self.handler.flat_wts

        # define iterables containing fit y & wts for each fit
        self.iter_flat = zip(self.observed_flat, self.flat_wts)
        self.iter_cond = zip(self.observed, self.cond_wts)
        # self.resultsdir = os.self.handler.make_results_dir(custompath=self.custompath, get_path=True)
        # get working directory
        self.resultsdir = os.path.abspath('./')


    def set_fitparams(self, force=None, **kwargs):
        """ dictionary of fit parameters, passed to Optimizer/Simulator objects
        """
        if not hasattr(self, 'fitparams'):

            # initialize with default values and first arrays in observed_flat, flat_wts
            #local methods = ['nelder', 'powell', 'lbfgsb' 'tnc', 'cobyla']
            self.fitparams = {'ix':0,
                            'ntrials': 10000,
                            'si': .1,
                            'dt':.003,
                            'tol': 1.e-25,
                            'method': 'nelder',
                            'maxfev': 800,
                            'maxiter': 800,
                            'kind': self.kind,
                            'clmap': self.clmap,
                            'pcmap':self.pcmap,
                            'depends_on': self.depends_on,
                            'ssd_method': self.handler.ssd_method,
                            'quantiles': self.quantiles,
                            'fit_on': self.fit_on,
                            'model_id': self.model_id,
                            'learn': self.learn,
                            'inits': self.inits,
                            'nlevels': 1,
                            'nidx': self.nidx,
                            'idx': self.idx[0],
                            'tb': self.tb}

            self.fitparams = pd.Series(self.fitparams)

        else:
            # fill with kwargs (i.e. y, wts, ix, etc)
            for kw_arg, kw_val in kwargs.items():
                self.fitparams[kw_arg] = kw_val

        if 'quantiles' in list(kwargs):
            self.update_quantiles()

        if 'depends_on' in list(kwargs):
            self.__prepare_fit__(kwargs['depends_on'])
            self.generate_model_id()
            self.set_fitparams(nlevels = self.nlevels, clmap=self.clmap, model_id=self.model_id)
            self.fitparams['nlevels'] = self.nlevels

        if force=='cond':
            self.fitparams['nlevels'] = self.nlevels

        elif force=='flat':
            self.fitparams['nlevels'] = 1

        self.update_data(self.fitparams.nlevels)

        if hasattr(self, 'ssdDF') and 'ssd' in self.data.columns:
            self.__set_ssd_info__()

        ksData=None
        if self.ksfit:
            ksData = self.get_ksdata(nlevels=self.fitparams['nlevels'])

        if hasattr(self, 'opt'):
            self.opt.update(fitparams=self.fitparams, inits=self.fitparams.inits, ksData=ksData)
            self.sim = self.opt.sim


    def set_basinparams(self, **kwargs):
        """ dictionary of global fit parameters, passed to Optimizer/Simulator objects
        """
        if not hasattr(self, 'basinparams'):
            self.basinparams =  {'ninits': 4,
                                'nsamples': 2500,
                                'interval': 10,
                                'T': .015,
                                'mutation': (0.5, 1),
                                'stepsize': .25,
                                'niter': 500,
                                'maxiter': 500,
                                'nsuccess': 150,
                                'polish_tol': 1.e-25,
                                'tol': .001,
                                'method': 'evolution',
                                'local_method': 'powell', #'L-BFGS-B', 'TNC'
                                'sample_method':'random',
                                'popsize': 15,
                                'recombination': .65,
                                'progress': True,
                                'strategy': 'best1bin',
                                'disp': False}
        else:
            # fill with kwargs for the upcoming fit
            for kw_arg, kw_val in kwargs.items():
                self.basinparams[kw_arg] = kw_val
            if self.basinparams['method']=='basin':
                self.basinparams['local_method']='TNC'
                self.basinparams['tol'] = 1e-20

        if hasattr(self, 'opt'):
            old_method = self.opt.basinparams['method']
            self.opt.update(basinparams=self.basinparams)
            self.sim = self.opt.sim
            if old_method != self.basinparams['method']:
                self.opt.make_progress_bars()


    def get_ksdata(self, nlevels=1):
        self.ksDataCond = {'corRT':[], 'errRT':[], 'goAcc':[], 'stopAcc':[]}
        self.ksDataFlat = deepcopy(self.ksDataCond)


        df = self.data.copy()
        self.ksDataFlat['corRT'].append(df[(df.ttype=='go')&(df.acc==1)].rt.values)
        self.ksDataFlat['errRT'].append(df[(df.ttype=='stop')&(df.acc==0)].rt.values)
        self.ksDataFlat['goAcc'].append(df[df.ttype=='go'].acc.mean())

        if self.ssd_method=='all':
            nssd = self.fitparams.ssd_info[1]
            self.ksDataFlat['stopAcc'].append(df.iloc[:, 3:3+nssd].mean())
        else:
            self.ksDataFlat['stopAcc'].append(df[df.ttype=='stop'].acc.mean())

        for col, vals in listitems(self.clmap):
            for lvl in vals:
                if lvl.isalnum():
                    try:
                        lvl = int(lvl)
                    except ValueError:
                        lvl = lvl
                dfi = self.data[self.data[col]==lvl]
                self.ksDataCond['corRT'].append(dfi[(dfi.ttype=='go')&(dfi.acc==1)].rt.values)
                self.ksDataCond['errRT'].append(dfi[(dfi.ttype=='stop')&(dfi.acc==0)].rt.values)
                self.ksDataCond['goAcc'].append(dfi[dfi.ttype=='go'].acc.mean())

                if self.ssd_method=='all':
                    dfStop = self.observedDF[self.observedDF[col]==lvl]
                    nssd = self.fitparams.ssd_info[1]
                    self.ksDataCond['stopAcc'].append(dfStop.iloc[:, 3:3+nssd].mean().values)
                else:
                    self.ksDataCond['stopAcc'].append(dfi[dfi.ttype=='stop'].acc.mean())

                #self.ksDataCond['stopAcc'].append(dfi[(dfi.ttype=='stop')&(dfi.probe==1)].acc.mean())

        if nlevels==1:
            return self.ksDataFlat
        return self.ksDataCond


    def update_data(self, nlevels=1):
        """ called when ix (int) is passed to fitparams as kwarg.
        Fills fitparams with y and wts vectors corresponding to ix'th
        arrays in observed(_flat) and (flat/cond)_wts lists.
        """
        i = self.fitparams['ix']
        if nlevels>1:
            self.fitparams['y'] = self.observed[i]
            self.fitparams['wts'] = self.cond_wts[i]
        else:
            self.fitparams['y'] = self.observed_flat[i]
            self.fitparams['wts'] = self.flat_wts[i]
        if self.fit_on=='subjects':
            self.fitparams['idx']=str(self.idx[i])
        else:
            self.fitparams['idx'] = 'avg'


    def sample_theta(self):

        pkeys = list(self.inits)
        nsamples = self.basinparams['nsamples']
        method = self.basinparams['sample_method']
        kind = self.kind

        if self.ssdelay:
            kind = '_'.join([kind, 'sso'])
        if self.gbase:
            kind = '_'.join([kind, 'z'])

        datadir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'inits')
        theta_fname = "{}_{}".format(method, kind)
        yhat_fname = "{}_{}_yhat".format(method, kind)
        thetaAll = pd.read_pickle(os.path.join(datadir, theta_fname), compression='xz')
        yhatAll = pd.read_pickle(os.path.join(datadir, yhat_fname), compression='xz')
        self.init_params = thetaAll.to_dict('records')
        groupedYhat = yhatAll.groupby('pset')
        yhats_ = groupedYhat.apply(analyze.rangl_data, self.ssd_method, self.quantiles)
        self.init_yhats = pd.DataFrame(np.vstack(yhats_.values))
        self.opt.init_params = self.init_params
        self.opt.init_yhats = self.init_yhats
        self.opt.sample_theta()


    def set_conditions(self, depends_on=None, bwfactors=None):
        data = self.data.copy()
        self.depends_on = depends_on
        self.conds = np.unique(np.hstack(listvalues(self.depends_on))).tolist()
        self.nconds = len(self.conds)
        if 'flat' in self.conds:
            self.is_flat = True
            data['flat'] = 'flat'
            self.data = data.copy()
        else:
            self.is_flat = False
        clevels = [np.sort(data[c].unique()) for c in self.conds]
        clevels = [np.array([str(lvl) for lvl in levels]) for levels in clevels]
        self.clmap = {c: lvls for c, lvls in zip(self.conds, clevels)}
        self.cond_matrix = np.array([lvls.size for lvls in clevels])
        self.nlevels = np.cumprod(self.cond_matrix)[-1]
        self.groups = np.hstack([['idx'], self.conds]).tolist()

        self.__format_pcmap__()
        if hasattr(self, 'ssdDF') and not self.is_nested:
            self.__set_ssd_info__()

        if hasattr(self, 'fitparams') and not self.is_nested:
            self.generate_model_id()
            self.set_fitparams(nlevels=self.nlevels, clmap=self.clmap, model_id=self.model_id)


    def __format_pcmap__(self):
        """ dict used by Simulator to extract conditional parameter values by name
        from lmfit Parameters object
            |<--- PARAMETERS OBJECT [LMFIT] <------- [IN]
            |---> p = {'v_bsl': V1, 'v_pnl': V2...} --->|
            |<--- pcmap = {'v':['v_bsl', 'v_pnl']} <---|
            |---> p['v'] = array([V1, V2]) -------> [OUT]
        """
        pcmap = {}
        for p, conds in self.depends_on.items():
            if isinstance(conds, list):
                levels = []
                for cond in conds:
                    levels.append(self.clmap[cond])
                level_data = list(product(*levels))
                clevels = ['_'.join([str(lvl) for lvl in lvls]) for lvls in level_data]
            else:
                clevels = [lvl for lvl in self.clmap[conds]]

            param_clevels = ['{}_{}'.format(p, clvl) for clvl in clevels]
            pcmap[p] = param_clevels

        self.pcmap = pcmap
        if hasattr(self, 'handler'):
            self.handler.pcmap = pcmap


    def __set_ssd_info__(self):
        """ set ssd_info for upcoming fit and store in fitparams dict
        """
        if self.fit_on=='average':
            ssdDF = self.ssdDF.copy()
            ssdDF = ssdDF.drop('idx', axis=1)
            ssd = ssdDF.groupby(self.conds).mean().values
        else:
            idx = self.idx[self.fitparams['ix']]
            # get ssd vector for fit index == ix
            ssd = self.ssdDF[self.ssdDF['idx'] == idx].groupby(self.conds).mean().values[:,1:]
            # if self.bwfactors is not None and hasattr(self, 'sim'):
            #     ix = self.conds.index([c for c in self.conds if c!=self.bwfactors][0])
            #     ssd = ssd[self.sim.pvary_ix[ix]]
            self.ssd = ssd
        if self.fitparams.nlevels==1:
            # single vector (nlevels=1), don't squeeze
            ssd = np.mean(ssd, axis=0, keepdims=True)
        nssd = ssd.shape[-1]
        nss = int((.5 * self.fitparams.ntrials))
        # nss = self.fitparams.ntrials
        nss_per_ssd = int(nss/nssd)
        ssd_ix = np.arange(nssd) * np.ones((ssd.shape[0], ssd.shape[-1])).astype(np.int)
        # store all ssd_info in fitparams, accessed by Simulator
        self.fitparams['ssd_info'] = [ssd, nssd, nss, nss_per_ssd, ssd_ix]


    def update_quantiles(self):
        """ recalculate observed dataframes w/ passed quantiles array
        """
        self.quantiles = self.fitparams.quantiles
        self.__make_dataframes__()
        self.fitparams['y'] = self.observed_flat[self.fitparams['ix']]
        self.fitparams['wts'] = self.flat_wts[self.fitparams['ix']]


    def set_results(self, finfo=None, popt=None, yhat=None):
        if finfo is None:
            finfo = self.finfo
        if popt is None:
            popt = self.popt
        if yhat is None:
            yhat = self.yhat
        return finfo, popt, yhat


    def generate_model_id(self, appendstr=None):
        """ generate an identifying string with model information.
        used for reading and writing model output
        """
        model_id = list(self.depends_on)
        if 'all' in model_id:
            model_id = ['flat']
        model_id.insert(0, self.kind)
        fit_on = 'avg'
        if self.fit_on=='subjects':
            fit_on = 'idx'
        model_id.append(fit_on)
        if appendstr is not None:
            model_id.append(appendstr)
        self.model_id = '_'.join(model_id)
        if hasattr(self, 'fitparams'):
            self.fitparams['model_id'] = self.model_id


    def set_testing_params(self, tol=1e-20, nsuccess=50, nsamples=1000, ninits=2, maxfev=1000, progress=True):
        self.set_fitparams(tol=tol, maxfev=maxfev)
        self.set_basinparams(tol=tol, ninits=ninits, nsamples=nsamples, nsuccess=nsuccess)
        self.opt.update(basinparams=self.basinparams, progress=progress)


    def toggle_pbars(self, progress=False, models=None):
        self.set_basinparams(progress=progress)
        if not progress:
            return None
        if self.fit_on=='subjects':
            status = ''.join(['Subj {}', '/{}'.format(self.nidx)])
            self.idxbar = utils.PBinJ(n=self.nidx, color='y', status=status)
        if models is not None:
            pvary = [list(depends_on) for depends_on in models]
            pnames = [vis.parameter_name(p, False) for p in pvary]
            self.mbar = utils.PBinJ(n=len(pnames), color='b', status='{} Model')
            return pnames


    def __remove_outliers__(self, sd=1.5, verbose=False):
        """ remove slow rts (>sd above mean) from main data DF
        """
        from radd.tools.analyze import remove_outliers
        self.data = analyze.remove_outliers(self.data.copy(), sd=sd, verbose=verbose)


    def __get_default_inits__(self):
        """ if inits not provided by user, initialize with default values
        see tools.theta.get_default_inits
        """
        self.inits = theta.get_default_inits(kind=self.kind, depends_on=self.depends_on, learn=self.learn, ssdelay=self.ssdelay, gbase=self.gbase)


    def __check_inits__(self, inits):
        """ ensure inits dict is appropriate for Model kind
        see tools.theta.check_inits
        """
        inits = dict(deepcopy(inits))
        checked = theta.check_inits(inits=inits, depends_on=self.depends_on, kind=self.kind)
        return checked



#!/usr/local/bin/env python
from __future__ import division
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
from radd import vis
from copy import deepcopy



def igt_scores(choices):
    A=len(choices[choices==0])
    B=len(choices[choices==1])
    C=len(choices[choices==2])
    D=len(choices[choices==3])
    # payoff (P) score
    P = (C+D) - (A+B)
    # sensitivity (Q) score
    Q = (B+D) - (A+C)
    return P, Q


def analyze_learning_dynamics(fd, broadcast=True, targets=['A', 'B', 'C', 'D']):

    choices = fd['choices']
    vd_all = fd['vd_all']
    vi_all = fd['vi_all']
    vdiff_all = vd_all - vi_all
    qdict = fd['qdict']
    nrows = len(choices)

    choice_vec = [targets[i] for i in choices]
    fd['choice']=choice_vec
    rts_copy = deepcopy(fd['rts'])
    fd['rt'] = [rts_copy[choice].pop(0) for i, choice in enumerate(choice_vec)]
    qcopy = deepcopy(qdict)
    fd['qval'] = [qcopy[choice].pop(0) for choice in choice_vec]

    fd['vd'] = [vd_all.loc[i, choice] for i, choice in enumerate(choice_vec)]
    fd['vi'] = [vi_all.loc[i, choice] for i, choice in enumerate(choice_vec)]
    fd['vdiff'] = [vdiff_all.loc[i, choice] for i, choice in enumerate(choice_vec)]
    fd['vdiff_all'] = [vdiff_all.loc[i, choice] for i, choice in enumerate(choice_vec)]
    vopt = vdiff_all['C'].values + vdiff_all['D'].values
    vsub = vdiff_all['A'].values + vdiff_all['B'].values

    vimp = vdiff_all['B'].values + vdiff_all['D'].values
    vnon = vdiff_all['A'].values + vdiff_all['C'].values
    fd['v_opt_diff'] = vopt - vsub
    fd['v_imp_diff'] = vimp - vnon
    #fd['agent'] = [fd['agent']*nrows]
    #q_go = fd['qdict_go']
    #q_no = fd['qdict_no']
    #qgo_copy = deepcopy(q_go)
    #qno_copy = deepcopy(q_no)
    #fd['q_go'] = [qgo_copy[choice].pop(0) for choice in choice_vec]
    #fd['q_no'] = [qno_copy[choice].pop(0) for choice in choice_vec]

    return fd

def format_dataframes(fd):
    from copy import deepcopy
    from collections import OrderedDict
    agdf_cols = ['agent', 'trial', 'agroup', 'qval', 'vd', 'vi', 'vdiff',
                 'v_opt_diff', 'v_imp_diff', 'choice', 'rt', 'a_go', 'a_no', 'adiff', 'beta']
    nrows = len(fd['choices'])
    fdbroad = dict(deepcopy(fd))
    #for key in ['agroup', 'agent', 'a_go', 'a_no', 'adiff', 'beta']:
    #    fdbroad[key] = [fd[key]]*nrows
    # agdf_cols = ['agent', 'trial', 'agroup', 'bgroup', 'group', 'qval', 'vd', 'vi', 'vdiff',
    #              'v_opt_diff', 'v_imp_diff', 'choice', 'rt', 'a_go', 'a_no', 'adiff', 'beta']
    agdf = pd.DataFrame(OrderedDict((col, fdbroad[col]) for col in agdf_cols))
    igtdf_cols=['agent', 'agroup', 'a_go', 'a_no', 'beta', 'P', 'Q']
    fd['P'], fd['Q'] = igt_scores(np.asarray(fd['choices']))

    igtdf = pd.Series(OrderedDict((col, fd[col]) for col in igtdf_cols))

    return igtdf, agdf


def calcPostErrAdjust(df, ntrials=200):
    if type(df.ttype.values[0]) is str:
        ssErrDF = df[(df.ttype=='stop')&(df.response==1)&(df.rt<.68)]
    else:
        ssErrDF = df[(df.ttype==0.)&(df.response==1)&(df.rt<.68)]
    ssErrRT = ssErrDF.rt.mean()
    adjustIX = ssErrDF.index.values + 1
    adjustIX[-1] = ssErrDF.index.values[-1]
    adjustDF = df.loc[adjustIX, :]
    PostErrRT = adjustDF[(adjustDF.response==1)&(adjustDF.rt<.68)].rt.mean()
    PostErrAdjust = PostErrRT - ssErrRT
    return PostErrAdjust

def calcTargetAdjust(df, ntrials=200):
    if type(df.ttype.values[0]) is str:
        GoDF = df[(df.ttype=='go')&(df.response==1)&(df.rt<.68)]
    else:
        GoDF = df[(df.ttype==1.)&(df.response==1)&(df.rt<.68)]
    GoRT = GoDF.rt.mean()
    adjustIX = GoDF.index.values + 1
    adjustIX[-1] = GoDF.index.values[-1]
    adjustDF = df.loc[adjustIX]
    GoAdjustRT = adjustDF[(adjustDF.response==1)&(adjustDF.rt<.68)].rt.mean()
    TargetAdjust =  GoAdjustRT - GoRT
    return TargetAdjust



from setuptools import setup, find_packages
import numpy as np
import os
package_data = {'radd':['docs/*.md', 'docs/*.txt', 'styles/*.css', 'docs/datasets/*', 'docs/examples/*.csv', 'docs/examples/*.mp4']}
setup(
    name='RADD',
    version='0.4.0',
    author='Kyle Dunovan, Timothy Verstynen, Jeremy Huang',
    author_email='dunovank@gmail.com',
    url='http://github.com/CoAxLab/radd',
    packages=['radd', 'radd.compiled', 'radd.adapt', 'radd.tools', 'radd.docs', 'radd.docs.examples'],
    package_data=package_data,
    description='RADD (Race Against Drift-Diffusion model) is a python package for fitting & simulating cognitive models of reinforcement learning and decision-making',
    install_requires=['numpy>=1.8.2', 'scipy>=0.16.1', 'matplotlib>=1.4.3', 'seaborn>=0.5.1', 'pandas>=0.15.1', 'lmfit>=0.9.1', 'scikit-learn>=0.17.1', 'progressbar2>=3.9.3', 'numba>=0.30.1', 'pyDOE', 'future'],
    include_dirs = [np.get_include()],
    classifiers=[
                'Environment :: Console',
                'Operating System :: OS Independent',
                'License :: OSI Approved :: BSD License',
                'Intended Audience :: Science/Research',
                'Development Status :: 3 - Alpha',
                'Programming Language :: Python',
                'Programming Language :: Python :: 2',
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.2',
                'Programming Language :: Python :: 3.4',
                'Programming Language :: Python :: 3.6',
                'Topic :: Scientific/Engineering',
                ]
)



#!usr/bin/env python
from __future__ import division
import pandas as pd
import numpy as np
import os
from numpy import array
from future.utils import listvalues
from sklearn.neighbors import KernelDensity
from scipy.stats.mstats import mquantiles as mq
from scipy.stats.mstats_extras import mjci
from scipy import optimize
import functools
from scipy.interpolate import interp1d
from scipy import stats
from scipy.optimize import leastsq, minimize, curve_fit


def ezdiff(rt, correct, s=1.):

    """ estimate ezDDM parameters
    """

    logit = lambda p:np.log(p/(1-p))
    pc = np.mean(correct)

    # subtract or add 1/2 an error to prevent division by zero
    if pc==1.0:
        pc=1 - 1/(2*len(correct))
    if pc==0.5:
        pc=0.5 + 1/(2*len(correct))
    MRT = np.mean(rt[correct==1])
    VRT = np.var(rt[correct==1])

    r = (logit(pc)*(((pc**2) * logit(pc)) - pc*logit(pc) + pc - 0.5))/VRT
    v = np.sign(pc-0.5)*s*(r)**0.25
    a = (s**2 * logit(pc))/v
    y = (-1*v*a)/(s**2)
    MDT = (a/(2*v))*((1-np.exp(y))/(1+np.exp(y)))
    tr = MRT-MDT

    return([a, v, tr])



def blockify_trials(data, nblocks=5, conds=None, groups=['idx']):

    datadf = data.copy()
    if conds is not None:
        if type(conds) is str:
            conds = [conds]
        groups = groups + conds

    idxdflist = []
    for dfinfo, idxdf in datadf.groupby(groups):
        ixblocks = np.array_split(idxdf.trial.values, nblocks)
        blocks = np.hstack([[i+1]*arr.size for i, arr in enumerate(ixblocks)])
        idxdf = idxdf.copy()
        colname = 'block{}'.format(nblocks)
        idxdf[colname] = blocks
        idxdflist.append(idxdf)

    return pd.concat(idxdflist)



def rangl_data(data, ssd_method='all', quantiles=np.linspace(.01,.99,15), fit_on='average'):
    """ called by DataHandler.__make_dataframes__ to generate
    observed data arrays
    """
    data = data.copy()
    if 1 in data.ttype.unique():
        goQuery = 'ttype==1.'
        stopQuery = 'ttype==0.'
    else:
        goQuery = 'ttype=="go"'
        stopQuery = 'ttype=="stop"'
    gac = data.query(goQuery).acc.mean()
    grt = data.query('response==1 & acc==1').rt.values
    ert = data.query('response==1 & acc==0').rt.values
    gq = mq(grt[grt<5.], prob=quantiles)
    eq = mq(ert[ert<5.], prob=quantiles)
    data_vector = [gac, gq, eq]
    if 'ssd' in data.columns:
        stopdf = data.query(stopQuery)
        if 'probe' in stopdf.columns:
            stopdf = stopdf[stopdf.probe==1]
        if ssd_method=='all':
            sacc=stopdf.groupby('ssd').mean()['acc'].values
        elif ssd_method=='central':
            sacc = np.array([stopdf.mean()['acc']])
        data_vector.insert(1, sacc)
    return np.hstack(data_vector)


def rangl_freq(df, quantiles=np.arange(.1, 1.,.2)):

    prob = np.asarray([0.] + quantiles.tolist() + [1.])
    nbins = quantiles.size+1

    godf = df[df.ttype=='go']
    ssdf = df[df.ttype=='stop']
    cdf = godf[godf.response==1]
    edf = ssdf[ssdf.response==1]

    qcor = mq(cdf.rt.values, prob)
    qerr = mq(edf.rt.values, prob)

    Ncor = cdf.shape[0]
    Nerr = edf.shape[0]
    prop_go = np.asarray([cdf[(cdf.rt >= qcor[i])&(cdf.rt < qcor[i+1])].shape[0] / Ncor for i in range(nbins)])
    prop_ss = np.asarray([edf[(edf.rt >= qerr[i])&(edf.rt < qerr[i+1])].shape[0] / Nerr for i in range(nbins)])

    O_cor = Ncor * prop_go
    O_err = Nerr * prop_ss
    return np.concatenate([O_cor, O_err])


def rangl_counts(df):
    Ncor = df[(df.ttype=='go')&(df.response==1)].shape[0]
    Nerr = df[(df.ttype=='stop')&(df.response==1)].shape[0]
    return np.hstack([Ncor, Nerr])


def calc_chi_square(rt, ssrt, fitparams):
    O_freq = fitparams['obs_freq']
    Nobs = fitparams['nobs']
    Ncor, Nerr = Nobs
    quantiles = fitparams['quantiles']
    ntrials = fitparams['ntrials']
    tb = fitparams['tb']
    y = fitparams['y']
    # rt, ssrt = m.sim.simulate_model(m.inits, analyze=False, get_rts=True)

    nbins = quantiles.size+1
    ssrt = ssrt.reshape(nl, -1)
    ert = rt[:, :int(ntrials/2)]

    Ncor_hat = np.asarray([rt_i[rt_i<tb].shape[0] for rt_i in rt])
    Nerr_hat= np.asarray([ert_i[ert_i<ssrt_i].shape[0] for ert_i, ssrt_i in zip(ert, ssrt)])

    qcor, qerr = np.split(y[-quantiles.size*2:], 2)
    qcor = np.hstack([[0.], qcor, [tb]])
    qerr = np.hstack([[0.], qerr, [tb]])

    prop_go = [[rti[(rti >= qcor[i])&(rti < qcor[i+1])].shape[0] / Ni for i in range(nbins)] for Ni, rti in zip(Ncor_hat, rt)]
    prop_ss = [[erti[(erti >= qerr[i])&(erti < qerr[i+1])].shape[0] / Ni for i in range(nbins)] for Ni, erti in zip(Nerr_hat, ert)]

    # Ncor, Nerr are actual counts of O_cor, O_err
    M_cor = [Ni * np.asarray(prop_go_i) for Ni, prop_ss_i in zip(Nerr, prop_go)]
    M_err = [Ni * np.asarray(prop_ss_i) for Ni, prop_ss_i in zip(Nerr, prop_ss)]
    M_freq = [np.concatenate([Mcor_i, Merr_i]) for Mcor_i, Merr_i in zip(M_cor, M_err)]

    chi2 = np.sum([np.sum((M_i - O_i)**2 / O_i) for M_i, O_i in zip(M_freq, O_freq)])

    return chi2


def calculate_qpdata(data, quantiles=np.linspace(.02,.98,10)):
    data = data[data.response==1].copy()
    cor = data[data['acc']==1]
    err = data[data['acc']==0]
    idx_qc = pd.DataFrame([mq(idxdf.rt.values, quantiles) for idx, idxdf in cor.groupby('idx')])
    idx_qe = pd.DataFrame([mq(idxdf.rt.values, quantiles) for idx, idxdf in err.groupby('idx')])
    quantMeans = np.vstack = ([idx_qc.mean(axis=0).values, idx_qe.mean(axis=0).values])
    quantErr = np.vstack([idx_qc.sem(axis=0).values, idx_qe.sem(axis=0).values])*1.96
    return quantMeans, quantErr


def idx_quant_weights_OLD(df, groups=[], nsplits=1, prob=np.arange(.1,1.,.1), max_wt=2.5):

    """ calculates weight vectors for reactive RT quantiles by
    first estimating the SEM of RT quantiles for corr. and err. responses.
    (using Maritz-Jarrett estimatation: scipy.stats.mstats_extras.mjci).
    Then representing these variances as ratios.
    e.g.
          QSEM = mjci(rtvectors)
          wts = median(QSEM)/QSEM
    """
    idx_mjci = lambda x: mjci(x.rt, prob=prob)

    nquant = prob.size

    if 'idxN' in df.columns:
        nidx = df.idxN.unique().size
    else:
        nidx = df.idx.unique().size

    # get all trials with response recorded
    godf = df.query('response==1')

    # sort by ttype first so go(acc==1) occurs before stop(acc==0)
    ttype_ordered_groups = np.hstack([groups, 'ttype', 'acc']).tolist()
    godf_grouped = godf.groupby(ttype_ordered_groups)

    # apply self.idx_mjci() to estimate quantile CI's
    quant_err = np.vstack(godf_grouped.apply(idx_mjci).values)

    # reshape [nidx   x   ncond * nquant * nacc]
    idx_qerr = quant_err.reshape(nidx, nsplits * nquant * 2)

    # calculate subject median across all conditions quantiles and accuracy
    # this implicitly accounts for n_obs as the mjci estimate of sem will be lower
    # for conditions with greater number of observations (i.e., more acc==1 in godf)
    idx_medians = np.nanmedian(idx_qerr, axis=1)
    idx_qratio = idx_medians[:, None] / idx_qerr

    # set extreme values to max_wt arg val
    idx_qratio[idx_qratio >= max_wt] = max_wt

    # reshape to fit in wtsDF[:, q_cols]
    return idx_qratio.reshape(nsplits * nidx, nquant * 2)


def idx_quant_weights(data, conds, quantiles=np.linspace(.02,.98,10), max_wt=3.5, bwfactors=None):
    """ calculates weight vectors for reactive RT quantiles by
    first estimating the SEM of RT quantiles for corr. and err. responses.
    (using Maritz-Jarrett estimatation: scipy.stats.mstats_extras.mjci).
    Then representing these variances as ratios.
    e.g.
          QSEM = mjci(rtvectors)
          wts = median(QSEM)/QSEM
    """
    idx_mjci = lambda x: mjci(x.rt, prob=quantiles)

    if bwfactors is not None:
        if not isinstance(bwfactors, list):
            bwfactors = [bwfactors]
        conds = [c for c in conds if c not in bwfactors]

    if len(conds)<1:
        nlevels = 1
    else:
        # get all trials with response recorded
        nlevels = np.sum([data[c].unique().size for c in conds])

    godf = data.query('response==1')
    # sort by ttype first so go(acc==1) occurs before stop(acc==0)
    ttype_ordered_groups = np.hstack([conds, 'ttype', 'acc']).tolist()
    godf_grouped = godf.groupby(ttype_ordered_groups)
    # apply self.idx_mjci() to estimate quantile CI's
    quant_err = np.hstack(godf_grouped.apply(idx_mjci).values)
    # calculate subject median across all conditions quantiles and accuracy
    # this implicitly accounts for n_obs as the mjci estimate of sem will be lower
    # for conditions with greater number of observations (i.e., more acc==1 in godf)
    idx_median = np.median(quant_err)
    # idx_qratio = idx_median / quant_err
    idx_qratio = idx_median / quant_err
    # set extreme values to max_wt arg val
    idx_qratio[idx_qratio >= max_wt] = max_wt

    return idx_qratio.reshape(nlevels, -1)


def idx_acc_weights(data, conds=['flat'], ssd_method='all'):
    """ count number of observed responses across levels, transform into ratios
    (counts_at_each_level / np.median(counts_at_each_level)) for weight
    subject-level p(response) values in cost function.
    ::Arguments::
        df (DataFrame): group-level dataframe
        var (str): column header for variable to count responses
        conds (list): depends_on.values()
    """
    df = data.copy()
    index=['idx']
    if not 'flat' in conds:
        index = index + conds
    if 'ssd' in df.columns:
        if ssd_method=='all':
            split_by = 'ssd'
            df = df[df.ttype=='stop'].copy()
            if 'probe' in df.columns:
                df = df[df.probe==1].copy()
        else:
            split_by = 'ttype'
    else:
        split_by = conds
        _ = index.remove(split_by)
    df['n'] = 1
    countdf = df.pivot_table('n', index=index, columns=split_by, aggfunc=np.sum)
    idx_pwts = countdf.values / countdf.median(axis=1).values[:, None]
    idx_pwts = np.ones_like(idx_pwts) #+ .5
    # idx_pwts = countdf.values / countdf.median(axis=0).values
    if ssd_method=='all':
        go_wts = np.ones(countdf.shape[0]) #* .5
        # idx_pwts[:, -5] = .5
        # idx_pwts[:, -4] = .5
        # idx_pwts[:, -3] = .5
        # idx_pwts[:, -2] = .5
        # idx_pwts[:, -1] = .5

        idx_pwts = np.concatenate((go_wts[:,None], idx_pwts), axis=1)

    return idx_pwts


def assess_fit(y, wts, yhat, nvary=5):
    # residual = wts * (yhat - y)
    # fp['yhat'] = (residual / wts) + y
    # fill finfo dict with goodness-of-fit info
    y = y.flatten()
    wts = wts.flatten()
    yhat = yhat.flatten()
    finfo = pd.Series()
    finfo['nvary'] = nvary
    finfo['chi'] = np.sum((wts*(yhat - y))**2)
    finfo['ndata'] = len(yhat)
    finfo['df'] = finfo.ndata - finfo.nvary
    finfo['rchi'] = finfo.chi / finfo.df
    finfo['logp'] = finfo.ndata * np.log(finfo.rchi)
    finfo['AIC'] = finfo.logp + 2 * finfo.nvary
    finfo['BIC'] = finfo.logp + np.log(finfo.ndata * finfo.nvary)
    return finfo


def pandaify_results(gort, ssrt, tb=.7, bootstrap=False, bootinfo={'nsubjects':25, 'ntrials':1000, 'groups':['ssd']}, ssd=np.array([[.2, .25, .3, .35, .4]]), clmap=None):
    nl, nssd, nssPer = ssrt.shape
    nl, ntrials = gort.shape
    nss = nssd * nssPer
    dfColumns=['cond', 'ttype', 'ssd', 'response', 'acc', 'rt', 'ssrt', 'trial']
    levelNames = np.hstack(listvalues(clmap)).tolist()
    # bootinfo['groups'] = bootinfo['groups'] + list(clmap)
    dfIndex = np.arange(ntrials)
    dfList = []
    for i in range(nl):
        ert = gort[i, :nss].reshape(ssrt[i].shape)
        goTrialOutcomes = np.hstack(np.where(gort[i, nss:] < tb, 1, 0))
        # ssTrialOutcomes = np.hstack(np.hstack(np.where(ert <= ssrt[i], 1, 0)))
        ssTrialOutcomes = np.hstack(np.hstack(np.where(ssrt[i] <= ert, 0, 1)))
        delays = np.append(np.hstack([[delay]*nssPer for delay in ssd[i]]), [1000]*nss)
        response = np.append(ssTrialOutcomes, goTrialOutcomes)
        responseTime = gort[i]
        ssResponseTime = np.append(np.hstack(np.hstack(ssrt[i])), [np.nan]*nss)
        ttype = np.append(np.zeros(nss), np.ones(nss))
        acc = np.where(ttype == response, 1, 0)
        cond = [levelNames[i]]*ntrials
        trial = np.arange(1, ntrials+1)
        dfData = [cond, ttype, delays, response, acc, responseTime, ssResponseTime, trial]
        df = pd.DataFrame(dict(zip(dfColumns, dfData)), index=dfIndex)
        dfList.append(df[dfColumns])
    df = pd.concat(dfList)
    df.reset_index(drop=True, inplace=True)
    # stopdf = resultsdf[resultsdf.ttype==0.]
    df.loc[(df.rt==1000.)&(df.ssrt==1000.)&(df.ttype==0.), 'response'] = 0
    df.loc[(df.rt==1000.)&(df.ssrt==1000.)&(df.ttype==0.), 'acc'] = 1

    if bootstrap:
        resultsdf = bootstrap_data(resultsdf, nsubjects=bootinfo['nsubjects'], n=bootinfo['ntrials'], groups=bootinfo['groups'])
    return resultsdf


def bootstrap_data(data, nsubjects=25, n=120, groups=['cond', 'ssd']):
    """ generates n resampled datasets using rwr()
    for bootstrapping model fits
    """
    df = data.copy()
    bootlist = []
    idxList = []
    if 'idx' in df.columns:
        df = df.drop('idx', axis=1)
    if n == None:
        n = len(df)
    # bootdfList = []
    # for level, levelDF in df.groupby(groups[0]):
    bootlist = []
    for idx in range(nsubjects):
        boots = df.reset_index(drop=True)
        orig_ix = np.asarray(boots.index[:])
        resampled_ix = rwr(orig_ix, get_index=True, n=n)
        bootdf = df.iloc[resampled_ix]
        bootdf = bootdf.copy()
        bootdf.insert(0, 'idx', idx)
        bootdf['trial'] = np.arange(bootdf.shape[0])
        bootlist.append(bootdf)
    # concatenate and return all resampled conditions
    bootdf = pd.concat(bootlist)
    bootdf.reset_index(drop=True, inplace=True)
    # bootdfList.append(bootdf)
    # bootdf = pd.concat(bootdfList)
    return bootdf


def rwr(X, get_index=False, n=None):
    """
    Modified from http://nbviewer.ipython.org/gist/aflaxman/6871948
    """
    if isinstance(X, pd.Series):
        X = X.copy()
        X.index = range(len(X.index))
    if n == None:
        n = len(X)
    resample_i = np.floor(np.random.rand(n) * len(X)).astype(int)
    X_resample = (X[resample_i])
    if get_index:
        return resample_i
    else:
        return X_resample


def estimate_timeboundary(data):
    goDF = data[data.response==1].copy()
    return np.int(np.round(goDF.rt.max(), 2)*1000) * .001


def determine_ssd_method(stopdf):
    ssd_n = [df.size for _, df in stopdf.groupby('ssd')]
    # test if equal # of trials per ssd & return ssd_n
    all_equal_counts = ssd_n[1:] == ssd_n[:-1]
    if all_equal_counts:
        return 'all'
    else:
        return'central'


def get_model_ssds(stopdf, conds, ssd_method='all', scale=.001, bwfactors=None):

    if bwfactors is not None:
        if not isinstance(bwfactors, list):
            bwfactors = [bwfactors]
        conds = [c for c in conds if c not in bwfactors]

    groups = conds + ['idx']
    if ssd_method == 'all':
        get_idx_ssd = lambda x: pd.Series(np.sort(x.ssd.unique().squeeze())) * scale
        ssdidx = stopdf.groupby(groups).apply(get_idx_ssd)
        ssdDF = pd.DataFrame(ssdidx).reset_index()
    elif ssd_method == 'central':
        get_idx_ssd = lambda x: x.ssd.mean() * scale
        ssdidx = stopdf.groupby(groups).apply(get_idx_ssd)
        ssdDF = ssdidx.reset_index()

    return ssdDF


def load_nested_popt_dictionaries(fitdf, params={}, nlevels=1):
    mparams = list(params)
    params_dict = {}

    for mid in fitdf.idx.values:
        freeparams = mid.split('_')
        constants = [p for p in mparams if p not in freeparams]
        select_params = fitdf[fitdf.idx==mid][mparams]
        pdict = select_params.to_dict('records')[0]
        for p in constants:
            pdict[p] = pdict[p] * np.ones(nlevels)
        params_dict[mid] = pdict

    return params_dict


def fill_nan_vals(coldata):
    coldata = coldata.copy()
    coldata[coldata.isnull()] = np.nanmean(coldata.values)
    return coldata


def remove_outliers(data, sd=1.5, verbose=False):
    df = data.copy()
    ssdf = df[df.response == 0]
    godf = df[df.response == 1]
    bound = godf.rt.std() * sd
    rmslow = godf[godf['rt'] < (godf.rt.mean() + bound)]
    clean_go = rmslow[rmslow['rt'] > (godf.rt.mean() - bound)]
    clean = pd.concat([clean_go, ssdf])
    if verbose:
        pct_removed = len(clean) * 1. / len(df)
        print("len(df): %i\nbound: %s \nlen(cleaned): %i\npercent removed: %.5f" % (len(df), str(bound), len(clean), pct_removed))
    sortby = 'idx'
    if 'trial' in clean.columns:
        sortby = ['idx', 'trial']
    cleandf = clean.sort_values(by=sortby)
    return cleandf


def ensure_numerical_wts(wts, flat_wts):
    # test inf
    wts[np.isinf(wts)] = np.median(wts[~np.isinf(wts)])
    flat_wts[np.isinf(flat_wts)] = np.median(flat_wts[~np.isinf(flat_wts)])
    # test nan
    wts[np.isnan(wts)] = np.median(wts[~np.isnan(wts)])
    flat_wts[np.isnan(flat_wts)] = np.median(flat_wts[~np.isnan(flat_wts)])
    return wts, flat_wts


def kde_fit_quantiles(rtquants, nsamples=1000, bw=.1):
    """ takes quantile estimates and fits cumulative density function
    returns samples to pass to sns.kdeplot()
    """
    kdefit = KernelDensity(kernel='gaussian', bandwidth=bw).fit(rtquants)
    samples = kdefit.sample(n_samples=nsamples).flatten()
    return samples


def scurve_interpolate(x, y, kind='cubic'):
    interpol_fx = interp1d(x, y, kind=kind)
    xsim = np.linspace(x[0], x[-1], 50, endpoint=True)
    ysim = interpol_fx(xsim)
    return xsim, ysim


def scurve_poly_fit(x, y, n=20):
    polysim = lambda p, pi, x, xi: p[pi]*x**xi
    ix = np.arange(n+1)
    poly_ix = zip(ix, ix[::-1])
    p = np.polyfit(x,y,n)
    xsim = np.linspace(x.min()-.05, x.max()+.08, 2500, endpoint=True)
    ysim = np.sum([polysim(p, pi, xsim, xi) for pi, xi in poly_ix],axis=0)
    return xsim, ysim


def fit_sigmoid(x, y):
    #x, y = scurve_interpolate(x, y)
    initParams = [np.mean(x), .8, .5]
    results = minimize(sigmoid, initParams, args=(x, y), method='Nelder-Mead')
    popt = results.x
    xSim = np.linspace(x.min()-.01, x.max()+.01, 1000, endpoint=True)
    ySim = 1. / (1. + np.exp(popt[0]*(xSim-popt[1])))
    return xSim, ySim


def sigmoid(params, xdata, ydata):
    k = params[0]
    x0 = params[1]
    sd = params[2]
    yPred = 1. / (1. + np.exp(k*(xdata-x0)))
    # Calculate negative log likelihood
    LL = -np.sum(stats.norm.logpdf(ydata, loc=yPred, scale=sd ))
    return(LL)


def logistic4(x, a, b, c, d):
    """4PL lgoistic equation."""
    return ((a-d)/(1.0+((x/c)**b))) + d


def logistic_residuals(p, y, x):
    """Deviations of data from fitted 4PL curve"""
    a,b,c,d = p
    err = y-logistic4(x, a, b, c, d)
    return err


def logistic_peval(x, p):
    """Evaluated value at x with current parameters."""
    a,b,c,d = p
    return logistic4(x, a, b, c, d)


def fit_logistic(x, y):
    # Initial guess for parameters
    p0 = [0., .5, 1, 1]
    popt = leastsq(logistic_residuals, p0, args=(y, x))
    xSim = np.linspace(x.min()-.1, x.max()+.1, 1000, endpoint=True)
    ySim = logistic_peval(xSim, popt[0])
    return xSim, ySim


def residuals(p, x, y):
    return y - sigmoid(p, x)


def resize(arr, lower=0.0, upper=1.0):
    arr = arr.copy()
    if lower > upper:
        lower, upper = upper, lower
    arr -= arr.min()
    arr *= (upper - lower) / arr.max()
    arr += lower
    return arr

def get_intersection(iter1, iter2):
    """ get the intersection of two iterables ("items in-common")
    """
    intersect_set = set(iter1).intersection(set(iter2))
    return ([i for i in intersect_set])



#!/usr/local/bin/env python
from __future__ import division
import os
from copy import deepcopy
import numpy as np
import pandas as pd
from numpy import array
from radd.models import Simulator
from radd.CORE import RADDCore
from radd import vis
from radd.tools import utils, analyze, messages
from radd.tools.analyze import pandaify_results, rangl_data
import multiprocessing as mp
import matplotlib.pyplot as plt
from IPython.display import clear_output


class Model(RADDCore):
    """ Main class for instantiating, fitting, and simulating models.
    Inherits from RADDCore parent class (see CORE module).
    ::Arguments::
        data (pandas DF):
            data frame with columns 'idx', 'rt', 'acc', 'ttype', 'response',
            <Condition Name> declared in depends_on values
        kind (str):
            declares model type ['dpm', 'irace', 'pro']
            append 'x' to front of model name to include a dynamic
            bias signal in model
        inits (dict):
            dictionary of parameters (v, a, tr, ssv, z) used to initialize model
        fit_on (str):
            set if model fits 'average', 'subjects', 'bootstrap' data
        depends_on (dict):
            set parameter dependencies on task conditions
            (ex. depends_on={'v': 'Condition'})
        weighted (bool):
            if True (default), perform fits using a weighted least-squares approach
        quantiles (array):
            set the RT quantiles used to fit model
    """

    def __init__(self, data=pd.DataFrame, kind='xdpm', inits=None, fit_on='average', depends_on={'all':'flat'}, weighted=True, ssd_method=None, learn=False, bwfactors=None, custompath=None, presample=False, ssdelay=False, gbase=False, quantiles=np.arange(.1, 1.,.1), ksfit=False):

        super(Model, self).__init__(data=data, inits=inits, fit_on=fit_on, depends_on=depends_on, kind=kind, quantiles=quantiles, weighted=weighted, ssd_method=ssd_method, learn=learn, bwfactors=bwfactors, custompath=custompath, presample=presample, ssdelay=ssdelay, gbase=gbase, ksfit=ksfit)

        groups = self.handler.groups
        bwcol = None
        if self.bwfactors is not None:
            groups += [self.bwfactors]
            self.bwcol = [df[self.bwfactors].unique()[0] for _, df in self.data.groupby('idx')]


    def optimize(self, plotfits=False, saveplot=False, saveresults=False, custompath=None, progress=True, get_results=False, powell=True, hop=False):
        """ Method to be used for accessing fitting methods in Optimizer class
        see Optimizer method optimize()
        ::Arguments::
            plotfits (bool):
                if True (default), plot model predictions over observed data
            saveplot (bool):
                if True (default is False), save plots to "~/<self.model_id>/"
            saveresults (bool):
                if True (default), save fitdf, yhatdf, and txt logs to "~/<self.model_id>/"
            saveobserved (bool):
                if True (default is False), save observedDF to "~/<self.model_id>/"
            custompath (str):
                path starting from any subdirectory of "~/" (e.g., home).
                all saved output will write to "~/<custompath>/<self.model_id>/"
            progress (bool):
                track progress across ninits and basinhopping
        """
        #self.toggle_pbars(progress=progress)
        self.custompath=custompath

        if self.fit_on == 'subjects':
            finfo, popt, yhat  = self.optimize_idx_params(self.idx)

            if plotfits:
                self.plot_model_idx_fits(save=saveplot)

        else:
            self.set_fitparams(force='flat', nlevels=1)
            finfo, popt, yhat = self.optimize_flat(get_results=True)

            if not self.is_flat:
                finfo, popt, yhat = self.optimize_conditional(popt, get_results=True, powell=powell, hop=hop)

            if plotfits:
                self.plot_model_fits(save=saveplot)

        if progress:
            if self.fit_on=='subjects':
                self.idxbar.clear()
            self.opt.ibar.clear()
            self.opt.gbar.clear()

        try:
            self.write_results(finfo=finfo, popt=popt, yhat=yhat)
        except Exception:
            pass

        if get_results:
            return finfo, popt, yhat


    def fit_multi_idx(self, nproc=1, progress=True):
        nruns = int(len(self.idx) / nproc)
        fits, popts, yhats = [], [], []
        self.toggle_pbars(progress=progress)
        self.idxbar.update(value=0, status=0)
        self.opt.make_progress_bars(inits=False, basin=True)

        # self.iohandler = ModelIO(fitparams=self.fitparams, mname=self.model_id)
        self.yhatdf = self.observedDF[self.observedDF.idx.isin(self.idx)].copy()
        datcols = self.yhatdf.loc[:, 'acc':].columns.tolist()

        def fit_idx_i(pos, idx, output):
            self.set_fitparams(ix=self.idx.index(idx), force='flat', nlevels=1)
            finfo, popt, yhat = self.optimize_flat(get_results=True)
            if not self.is_flat:
                finfo, popt, yhat = self.optimize_conditional(popt, get_results=True)
            output.put((pos, {'finfo': finfo, 'popt': popt, 'yhat': yhat}))

        for i in range(nruns):
            # initiate process workers
            output = mp.Queue()
            ix0 = i * nproc
            ix1 = ix0 + nproc
            idxlist = self.idx[ix0:ix1]
            processes = [mp.Process(target=fit_idx_i, args=(pos, idx, output)) for pos, idx in enumerate(idxlist)]
            for p in processes: p.start()
            for p in processes: p.join()

            results = [output.get() for p in processes]; results.sort()
            results = np.hstack([r[1] for r in results]).tolist()
            fits.append(pd.concat([res['finfo'] for res in results], axis=1).T)
            popts.append(pd.DataFrame([res['popt'] for res in results]))
            yhat = np.array([res['yhat'] for res in results])
            self.yhatdf.loc[ix0:ix1, datcols] = yhat.reshape(nproc, -1)
            self.idxbar.update(value=ix1, status=ix1)

        self.fitdf = pd.concat(fits)
        self.poptdf = pd.concat(popts)
        self.poptdf.insert(0, 'idx', self.idx)
        if self.bwfactors is not None:
            self.poptdf.insert(1, self.bwfactors, self.bwcol)
            self.fitdf.insert(1, self.bwfactors, self.bwcol)
        self.idxbar.clear()
        return self.fitdf, self.poptdf, self.yhatdf



    def optimize_flat(self, param_sets=None, get_results=False):
        """ optimizes flat model to data collapsing across all conditions
        ::Arguments::
            None
        ::Returns::
            None
        ::Attributes Created::
            yhat_flat (array): model-predicted data array (ndim = self.levels)
            finfo_flat (pd.Series): fit info (AIC, BIC, chi2, redchi, etc)
            popt_flat (dict): optimized parameters dictionary
        """
        if self.basinparams['method']=='basin':
            # Global Optimization w/ Basinhopping (+TNC)
            gpopt = self.opt.hop_around(param_sets, learn=self.learn)
        else:
            # Global Optimization w/ Diff. Evolution (+TNC)
            gpopt, gfmin = self.opt.optimize_global(self.inits, learn=self.learn, resetProgress=False)

        # Flat Simplex Optimization of Parameters at Global Minimum
        self.set_fitparams(inits=gpopt)
        self.finfo, gpopt, self.yhat = self.opt.gradient_descent(p=gpopt, learn=self.learn)
        self.gpopt = deepcopy(gpopt)
        self.popt = deepcopy(gpopt)
        # if self.is_flat:
        #     self.write_results()
        if get_results:
            return self.finfo, self.popt, self.yhat
        if self.opt.progress:
            self.opt.ibar.clear()
            self.opt.gbar.clear()
            clear_output()
        return self.popt


    def optimize_conditional(self, flatp, hop=False, powell=True, get_results=False):
        """ optimizes full model to all conditions in data
        ::Arguments::
            None
        ::Returns::
            None
        ::Attributes Created::
            yhat (array): model-predicted data array (ndim = self.levels)
            finfo (pd.Series): fit info (AIC, BIC, chi2, redchi, etc)
            popt (dict): optimized parameters dictionary
            flat_popt (dict): deepcopy of popt
        """
        gpopt = self.__check_inits__(deepcopy(flatp))
        self.set_fitparams(force='cond', inits=gpopt)
        # Pretune Conditional Parameters
        if hop:
            gpopt, fmin = self.opt.optimize_global(gpopt, learn=self.learn)
            self.opt.update(inits=gpopt)

        # Final Simplex Optimization(s)
        self.set_fitparams(method='nelder', maxfev=300)
        finfo, popt, yhat = self.opt.gradient_descent(p=gpopt, learn=self.learn)

        if powell:
            self.set_fitparams(method='powell', inits=popt, force='cond', maxfev=1000)
            self.opt.update(inits=popt)
            finfo, popt, yhat = self.opt.gradient_descent(p=popt, learn=self.learn)

        self.finfo, self.popt, self.yhat = finfo, popt, yhat

        if get_results:
            return self.finfo, self.popt, self.yhat


    def nested_optimize(self, depends=[], flatp=None, saveplot=True, plotfits=False, custompath=None, progress=True, saveresults=True, saveobserved=False):
        """ optimize a series of models using same init parameters where the i'th model
            has depends_on = {<depends[i]> : <cond>}.
            NOTE: only for models with fit_on='average'
        ::Arguments::
            depends (list):
                list of depends_on dictionaries to fit using a single set of init parameters (self.flat_popt)
            plotfits (bool):
                if True (default), plot model predictions over observed data
            saveplot (bool):
                if True (default), save plots to model.handler.results_dir
            custompath (str):
                path starting from any subdirectory of "~/" (e.g., home).
                all saved output will write to "~/<custompath>/<self.model_id>/"
            progress (bool):
                track progress across model fits, ninits, and basinhopping
        """

        if flatp is None:
            self.set_basinparams(method='basin', nsamples=1000)
            self.set_testing_params()
            self.set_fitparams(force='flat')
            flatp = deepcopy(self.optimize_flat())

        self.fitdf, self.poptdf, self.yhatdf = nested_optimize(depends, self.data, kind=self.kind, flatp=flatp, saveplot=saveplot, plotfits=plotfits, custompath=custompath, progress=progress, saveresults=saveresults)

        return self.fitdf, self.poptdf, self.yhatdf


    def log_fit_info(self, finfo=None, popt=None, yhat=None):
        """ write meta-information about latest fit
        to logfile (.txt) in working directory
        """
        self.opt.log_fit_info(finfo, popt, yhat)


    def write_results(self, finfo=None, popt=None, yhat=None):
        """ logs fit info to txtfile, fills yhatdf and fitdf
        """

        from radd import theta

        finfo, popt, yhat = self.set_results(finfo, popt, yhat)
        self.log_fit_info(finfo, popt, yhat)

        self.yhatdf = self.handler.fill_yhatdf(yhat=yhat, fitparams=self.fitparams)
        self.fitdf = self.handler.fill_fitdf(finfo=finfo, fitparams=self.fitparams)

        # pcopy = theta.scalarize_params(deepcopy(popt))
        pcopy = deepcopy(popt)
        self.poptdf = self.handler.fill_poptdf(popt=pcopy, fitparams=self.fitparams)



    def plot_model_fits(self, y=None, yhat=None, kde=True, err=None, save=False, bw='scott', savestr=None, same_axis=True, clrs=None, lbls=None, cumulative=True, simdf=None, suppressLegend=False, simData=None, condData=None, shade=True, plot_error_rts=True, figure=None, reorder=None):
        """ wrapper for radd.tools.vis.plot_model_fits
        """
        data = self.handler.data.copy()

        if y is None:
            y = self.fitparams.y

        if yhat is None:
            try:
                yhat = self.yhat
            except AttributeError:
                print("No Model Predictions to Plot (need yhat argument)")
                yhat = analyze.rangl_data(simdf, quantiles=self.quantiles)

        if reorder is not None:
            y = y.reshape(self.fitparams.nlevels, -1)
            yhat = yhat.reshape(self.fitparams.nlevels, -1)
            y = np.array([y[i] for i in reorder])
            yhat = np.array([yhat[i] for i in reorder])
        if save:
            if savestr is None:
                savestr = self.fitparams.model_id
            if self.fitparams['fit_on']=='subjects':
                savestr = savestr + str(self.fitparams['idx'])

        if err is None:
            err = self.observedErr.loc[:, 'acc':].values.squeeze()

        if lbls is None and self.fitparams.nlevels>1:
            from itertools import product
            levels = [self.clmap[cond] for cond in self.conds]
            level_data = list(product(*levels))
            lbls = ['_'.join([str(lvl) for lvl in lvls]) for lvls in level_data]

        if self.ssd_method == 'central':
            ssd = self.ssdDF.groupby(self.conds).mean()[0].values
            ssderr = self.ssdDF.groupby(self.conds).sem()[0].values
        else:
            ssd = self.fitparams.ssd_info[0]
            ssderr = None

        fig = vis.plot_model_fits(y, yhat, err=err, quantiles=self.quantiles, ssd=ssd, ssderr=ssderr, bw=bw, same_axis=same_axis, clrs=clrs, lbls=lbls, cumulative=cumulative, save=save, savestr=savestr, suppressLegend=suppressLegend, shade=shade, plot_error_rts=plot_error_rts, figure=figure)


    def simulate(self, p=None, analyze=True):
        """ simulate yhat vector using
        :: Arguments ::
            p (dict):
                parameters dictionary
            analyze (bool):
                if True (default) returns yhat vector. else, returns decision traces
        :: Returns ::
            out (array):
                1d array if analyze is True, else ndarray of decision traces
        """
        if p is None:
            try:
                p = self.popt
            except Exception:
                p = self.__get_default_inits()
        p = deepcopy(p)
        yhat = self.sim.simulate_model(p, analyze=analyze)
        return yhat


    def optimize_idx_params(self, idxlist=None, pos=0, output=None, progress=True):
        """ optimize parameters for individual subjects, store results
        :: Arguments ::
            param_sets (list):
                parameters dictionary
            force (str):
                if 'cond' forces fits to conditional data, if 'flat' forces flat data
            save (bool):
                save output dataframes if True
        :: Returns ::
            fitdf (DataFrame): fit statistics
            poptdf (DataFrame): optimized parameters
            yhatdf (DataFrame): model predictions
        """

        self.iohandler = ModelIO(fitparams=self.fitparams, mname=self.model_id)
        yhatdf = self.observedDF[self.observedDF.idx.isin(self.idx)].copy()
        datcols = yhatdf.loc[:, 'acc':].columns.tolist()
        # fit result lists for each param set
        finfoList, poptList, yhatList = [], [], []
        for idx in self.idx:

            self.toggle_pbars(progress=progress)

            ix = self.idx.index(idx)
            if hasattr(self, 'idxbar'):
                self.idxbar.update(value=ix, status=ix+1)

            # set subject data & wts
            self.set_fitparams(ix=ix, force='flat', nlevels=1)
            # optimize constants (flat model)
            finfo, popt, yhat = self.optimize_flat(get_results=True)
            flatPopt = {p: popt[p] for p in list(self.sim.inits)}
            # optimize conditional parameters
            if not self.is_flat:
                self.set_fitparams(force='cond')
                finfo, popt, yhat = self.optimize_conditional(popt, get_results=True)
                yhat = yhat.reshape(self.nlevels, -1)
                for p in list(self.depends_on):
                    popt[p] = flatPopt[p]

            # store results
            finfoList.append(finfo)
            poptList.append(deepcopy(popt))
            yhatdf.loc[yhatdf.idx==idx, datcols] = yhat

            self.opt.ibar.clear()
            self.opt.gbar.clear()
            clear_output()

        # concatenate all subjects together into single fitdf, poptdf, & yhatdf
        fitdf = pd.concat(finfoList, axis=1).T
        poptdf = pd.DataFrame.from_dict(poptList)
        poptdf.insert(0, 'idx', self.idx)
        if self.bwfactors is not None:
            poptdf.insert(1, self.bwfactors, self.bwcol)
            fitdf.insert(1, self.bwfactors, self.bwcol)

        self.iohandler.save_model_results(fitdf, poptdf, yhatdf, write=True)
        self.fitdf, self.poptdf, self.yhatdf = self.iohandler.read_model_results()
        return self.fitdf, self.poptdf, self.yhatdf


def nested_optimize(depends, data, kind='xdpm', flatp=None, saveplot=True, plotfits=False, custompath=None, progress=True, saveresults=False, saveobserved=False):
    """ optimize a series of models using same init parameters where the i'th model
        has depends_on = {<depends[i]> : <cond>}.
        NOTE: only for models with fit_on='average'
    ::Arguments::
        depends (list):
            list of depends_on dictionaries to fit using a single set of init parameters (m.flat_popt)
        data (DataFrame):
            input dataframe to fit nested models to
        plotfits (bool):
            if True (default), plot model predictions over observed data
        saveplot (bool):
            if True (default), save plots to model.handler.results_dir
        custompath (str):
            path starting from any subdirectory of "~/" (e.g., home).
            all saved output will write to "~/<custompath>/<m.model_id>/"
        progress (bool):
            track progress across model fits, ninits, and basinhopping
    """

    fits, popts, yhats = [], [], []

    for i, depends_on in enumerate(depends):
        m = Model(data=data, kind=kind, depends_on=depends_on)
        m.set_basinparams(method='basin', nsamples=1000)
        m.set_testing_params()
        m.custompath = custompath
        pnames = m.toggle_pbars(progress=progress, models=depends)

        if progress:
            m.mbar.update(value=i, status=pnames[i])

        m.set_fitparams(force='cond')
        flatp_ = deepcopy(flatp)
        finfo, popt, yhat = m.optimize_conditional(flatp_, get_results=True)

        m.write_results(finfo, popt, yhat)
        m.fitdf.insert(0, 'modelID', '_'.join(list(m.depends_on)))
        m.poptdf.insert(0, 'modelID', '_'.join(list(m.depends_on)))
        m.yhatdf.insert(0, 'modelID', '_'.join(list(m.depends_on)))
        fits.append(m.fitdf); popts.append(m.poptdf); yhats.append(m.yhatdf)

        if plotfits:
            m.plot_model_fits(save=saveplot)

        if progress:
            m.mbar.clear()
            m.opt.gbar.clear()
            m.opt.ibar.clear()

    return [pd.concat(outdata).reset_index(drop=True) for outdata in [fits, popts, yhats]]


class ModelIO(object):
    """ generates model read and write paths for
    handling I/O of model DataFrames, figures, etc
    """

    def __init__(self, fitparams, mname='radd_model', savestr=None, idxdir='subj_fits'):

        self.fitparams = fitparams
        self.idx = self.fitparams.idx
        self.nidx = len(self.idx)
        self.mname = mname

        if savestr==None:
            savestr = mname
        self.savestr = savestr
        if idxdir is not None:
            idxdir = os.path.join(os.path.expanduser('~'), idxdir)
        self.idxdir = idxdir
        self.mdir = os.path.join(self.idxdir, mname)

        makePath = lambda dfname: os.path.join(self.mdir, '_'.join([self.mname, dfname])+'.csv')
        dfnames = ['fitdf', 'poptdf', 'yhatdf']
        self.fitPath, self.poptPath, self.yhatPath = [makePath(dfname) for dfname in dfnames]
        self.paths = [self.idxdir, self.mdir]


    def save_model_results(self, fitdf=None, poptdf=None, yhatdf=None, write=True):
        """ save model fitdf, poptdf, and yhatdf dataframes
        """

        if fitdf is not None:
            self.fitdf = fitdf
        if poptdf is not None:
            self.poptdf = poptdf
        if yhatdf is not None:
            self.yhatdf = yhatdf
        if write:
            for pth in self.paths:
                if not os.path.isdir(pth):
                    os.mkdir(pth)
            self.fitdf.to_csv(self.fitPath, index=False)
            self.poptdf.to_csv(self.poptPath, index=False)
            self.yhatdf.to_csv(self.yhatPath, index=False)


    def read_model_results(self):
        """ save model fitdf, poptdf, and yhatdf dataframes
        """
        self.fitdf = pd.read_csv(self.fitPath)
        self.poptdf = pd.read_csv(self.poptPath)
        self.yhatdf = pd.read_csv(self.yhatPath)
        return self.fitdf, self.poptdf, self.yhatdf


    def save_fit_figure(self, f, savestr='avgYhat'):
        """ save model fits figure
        """
        savepath = os.path.join(self.mdir, '_'.join([self.mname, savestr]) + '.png')
        plt.tight_layout()
        f.savefig(savepath, dpi=600)



#!/usr/local/bin/env python
from __future__ import division
import os
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
from radd.IGT import visr, analyzer
from copy import deepcopy
from radd import theta
from scipy.stats.mstats import mquantiles as mq

temporal_dynamics = lambda p, t: np.cosh(p['xb'][:, na] * t)
updateQ = lambda q, winner, r, A: q[winner][-1] + A*(r - q[winner][-1])
softmax_update = lambda q, name, B: np.exp(B*q[name][-1])/np.sum([np.exp(B*q[k][-1]) for k in q.keys()])
go_resp = lambda trace, upper, dt: np.argmax((trace.T >= upper).T, axis=0) * dt
ss_resp = lambda trace, x, dt: np.argmax((trace.T <= 0).T, axis=0) * dt
go_RT = lambda ontime, rbool: ontime[:, na] + (rbool*np.where(rbool==0., np.nan, 1))
ss_RT = lambda ontime, rbool: ontime[:, :, na] + (rbool*np.where(rbool==0., np.nan, 1))
# RTQ = lambda zpd, prob: map((lambda x: mq(x[0][x[0] < x[1]], prob)), zpd)
RTQ = lambda zpd: [mquantiles(rt[rt < deadline], prob) for rt, deadline in zpd]

def run_full_sims(p, env=pd.DataFrame, alphas_go=[], alphas_no=None, betas=[], a_Q=.1, nblocks=2, nagents=30, si=.01, agent_list=[]):

    n_a = len(alphas_go)
    n_b = len(betas)
    trials = np.arange(1, (len(env)*nblocks)+1)

    if alphas_no is None:
        alphas_no=deepcopy(alphas_go)

    agroups = np.hstack([np.hstack([np.arange(n_a)]*n_b)]*nagents)
    bgroups = np.sort(np.hstack([np.sort(np.hstack([np.arange(n_b)]*n_a))]*nagents))
    agents = np.sort(np.hstack([np.arange(nagents)]*n_a*n_b))
    group = 0
    for agroup, bgroup, agent_i in zip(agroups, bgroups, agents):

        beta = betas[bgroup]
        a_go = alphas_go[agroup]
        a_no = alphas_no[agroup]

        pcopy=deepcopy(p)
        #p_rand = theta.random_inits(pcopy.keys())
        sim_out = run_trials(pcopy, env, nblocks=nblocks, si=si, a_go=a_go, a_no=a_no, beta=beta, a_Q=a_Q)
        choices, rts, all_traces, qdict, choicep, vd_all, vi_all = sim_out

        format_dict = {'agent': agent_i+1, 'trial': trials, 'a_go': a_go, 'a_no': a_no,
        'adiff': a_go-a_no, 'choices':choices, 'rts': rts, 'group': group, 'agroup': agroup+1,
        'bgroup': bgroup+1, 'qdict': qdict, 'choicep':choicep, 'vd_all':vd_all, 'vi_all':vi_all,
        'beta': beta}

        format_dict_updated = analyzer.analyze_learning_dynamics(format_dict)
        igtdf, agdf = analyzer.format_dataframes(format_dict_updated)
        agent_list.append([agdf, igtdf])
        group += 1

    trial_df = pd.concat([ag[0] for ag in agent_list]) #.groupby(['group', 'trial']).mean().reset_index()
    igt_df = pd.concat([ag[1] for ag in agent_list], axis=1).T #.groupby('group').mean().reset_index()

    return [trial_df, igt_df]

def vectorize_params(p, pcmap={'vd': ['vd_a', 'vd_b', 'vd_c', 'vd_d'], 'vi': ['vi_a', 'vi_b', 'vi_c', 'vi_d']}, nresp=4, sstrial=False):
    constants = ['a', 'tr', 'vd', 'vi', 'xb']
    if sstrial:
        constants.append('ssv')
    #remove conditional parameters from constants
    [constants.remove(param) for param in pcmap.keys()]
    for pkey in constants:
        p[pkey]=p[pkey]*np.ones(nresp)
    for pkey, pkc in pcmap.items():
        if nresp==1:
            break
        elif pkc[0] not in p.keys():
            p[pkey] = p[pkey]*np.ones(len(pkc))
        else:
            p[pkey] = array([p[pc] for pc in pkc])
    return p


def rew_func(rprob):
    if rs()<rprob:
        return 1
    else:
        return 0


def run_trials(p=None, cards=None, nblocks=2, si=.1, a_go=.2, a_no=.2, beta=5, aX=.001, dt=.001, plot=False, tb=1.1):
    """simulate series of trials with learning
    Arguments:
        p (dict): parameter dictionary
        cards (DataFrame): pandas DF (ntrials x nalt) with choice outcome vaulues
    Returns:
        choices (list): choice made on each trial
        rts (dict): rt for each trial (winner rt)
        all_traces (list): execution process traces truncated to length of winner
        qdict (dict): sequence of Q-value updates for each alt
    """
    if p is None:
        p={'vd':.09, 'vi':.05}; p['a']=.006; p['tr']=.3; p['xb']=.001
    trials = cards.copy()
    trials = trials.append([trials]*(nblocks-1)).reset_index()
    trials.rename(columns={'index':'t'}, inplace=True)
    ntrials=len(trials)
    choices, all_traces = [], []
    names = np.sort(cards.columns.values)
    rts={k:[] for k in names}
    qdict={k:[0] for k in names}
    choice_prob={k:[.25] for k in names}

    vdhist = pd.DataFrame(data=np.zeros((ntrials, len(names))), columns=names, index=np.arange(ntrials))
    vihist = vdhist.copy()
    needsfixed = 0
    prevWinner=np.nan
    for i in range(ntrials):
        vals = trials.iloc[i, 1:].values
        iquit=0
        winner=np.nan
        while np.isnan(winner) and iquit<35:
            execution = simulate_multirace(p, si=si, dt=dt)
            winner, rt, traces, p, qdict, choice_prob, choices = analyze_multiresponse(execution, p, qdict=qdict, vals=vals, names=names, a_go=a_go, a_no=a_no, aX=aX, beta=beta, choice_prob=choice_prob, dt=dt, choices=choices, prevWinner=prevWinner)
            iquit+=1
            if np.isnan(np.mean(p['xb'])):
                iquit=36
        prevWinner=winner
        if winner>=len(names) or np.isnan(winner):
            needsfixed+=1
            winner = int(np.random.choice(np.arange(len(names))))
        choice_name = names[choices[i]]
        oldval = trials.loc[i, choice_name]
        new_col = trials[choice_name].shift(-1)
        new_col.set_value(new_col.index[-1], oldval)
        trials=trials.copy()
        trials.loc[:, choice_name] = new_col

        vdhist.iloc[i, :] = p['vd']
        vihist.iloc[i, :] = p['vi']
        choice_name = names[winner]
        rts[choice_name].append(rt[winner]); all_traces.append(traces)
    percent_random_choice = (needsfixed*100.)/ntrials
    if percent_random_choice>=10.:
        print("trials with no winner {:.2f}%".format(percent_random_choice))
    if plot:
        visr.plot_traces_rts(p, all_traces, rts)
        return all_traces, rts
    return choices, rts, all_traces, qdict, choice_prob, vdhist, vihist


def simulate_multirace(p, pcmap={'vd': ['vd_a', 'vd_b', 'vd_c', 'vd_d'], 'vi': ['vi_a', 'vi_b', 'vi_c', 'vi_d']}, dt=.001, si=.01, tb=.9, single_process=0, return_di=False):

    nresp = len(pcmap.values()[0])
    dx = si * np.sqrt(dt)
    p = vectorize_params(p, pcmap=pcmap, nresp=nresp)

    Tex = np.ceil((tb-p['tr'])/dt).astype(int)
    xtb = temporal_dynamics(p, np.cumsum([dt]*Tex.max()))

    if single_process:
        Pe = 0.5*(1 + (p['vd']-p['vi'])*dx/si)
        execution = xtb * np.cumsum(np.where((rs((nresp, Tex.max())).T < Pe), dx, -dx).T, axis=1)
    else:
        Pd = 0.5 * (1 + p['vd'] * dx / si)
        Pi = 0.5 * (1 + p['vi'] * dx / si)
        direct = xtb * np.where((rs((nresp, Tex.max())).T < Pd),dx,-dx).T
        indirect = np.where((rs((nresp, Tex.max())).T < Pi),dx,-dx).T
        execution = np.cumsum(direct-indirect, axis=1)
        if return_di:
            return np.cumsum(direct, axis=1), np.cumsum(indirect, axis=1), execution
    return execution

def analyze_multiresponse(execution, p, qdict={}, vals=[], names=[], a_go=.2, a_no=.2, dt=.001, beta=5, choice_prob={}, aX=.001, choices=[], prevWinner=np.nan):
    """analyze multi-race execution processes"""
    nsteps_to_rt = np.argmax((execution.T>=p['a']).T, axis=1)
    rts = p['tr'] + nsteps_to_rt*dt
    # set non responses to 999
    rts[rts==p['tr'][0]]=999
    if np.all(rts==999):
        # if no response occurs, increase exponential bias (up to 3.0)
        if np.mean(p['xb']) <= 4.0:
            p['xb']=p['xb']*1.005
        return np.nan, rts, execution, p, qdict, choice_prob, choices
    # get accumulator with fastest RT (winner) in each cond
    winner = np.argmin(rts)
    choices.append(winner)
    # get rt of winner in each cond
    winrt = rts[winner]
    # slice all traces at time the winner crossed boundary
    traces = [execution[i, :nsteps_to_rt[winner]] for i in range(len(rts))]
    reward = vals[winner]
    winner_name = names[winner]
    if prevWinner==winner and reward>0:
        #p['xb'][winner] = p['xb'][winner] + aX * reward
        p['a'][winner] = p['a'][winner]*(1. - aX)
    elif prevWinner==winner and reward<0:
        p['a'][winner] = p['a'][winner]*(1. + aX)
        #p['xb'][winner] = p['xb'][winner] + aX * reward
    loser_names = names[names!=winner_name]
    # update action value
    qval = qdict[names[winner]][-1]
    if reward>=qval:
        alpha=a_go
    else:
        alpha=a_no
    if reward<0:
        reward
    Qt = updateQ(qdict, winner_name, reward, alpha)#, a_Q)
    qdict[winner_name].append(Qt)
    for lname in loser_names:
        qdict[lname].append(qdict[lname][-1])

    for alt_i, name in enumerate(names):
        cp_old = choice_prob[name][-1]
        # update choice probability using boltzmann eq. w/ inv. temp beta
        cp_new = softmax_update(qdict, name, beta)
        choice_prob[name].append(cp_new)
        # calc. change in choice probability for alt_i
        delta_prob = cp_new - cp_old
        # update direct & indirect drift-rates with cp_delta
        p = reweight_drift(p, alt_i, delta_prob, a_go, a_no)
        #p = weight_drift(p, alt_i, deltaQ, a_go, a_no)
    #p['a'] = array([a_no*(bound_expected-np.sum(p['vi']))]*p['a'].size)
    return winner, rts, traces, p, qdict, choice_prob, choices

def weight_drift(p, alt_i, deltaQ, a_go, a_no):
    vd_exp = p['vd'][alt_i]
    vi_exp = p['vi'][alt_i]
    p['vd'][alt_i] = vd_exp + a_go * (vd_exp - vd_exp * np.exp(1-(deltaQ)))
    p['vi'][alt_i] = vi_exp + a_no * (vi_exp - vi_exp * np.exp(1-(deltaQ)))
    return p

def reweight_drift(p, alt_i, delta_prob, a_go, a_no):
    """ update direct & indirect drift-rates for multirace winner
    """

    #p['vd'][alt_i] = vd_exp * np.exp(delta_prob)
    vd_exp = p['vd'][alt_i]
    #p['vd'][alt_i] = vd_exp + a_go * (vd_exp - (vd_exp * (1+delta_prob*2)))
    # p['vd'][alt_i] = vd_exp + a_go * (vd_exp - (vd_exp * np.exp(-delta_prob)))
    vi_exp = p['vi'][alt_i]
    #p['vi'][alt_i] = vd_exp + a_go * (vd_exp - (vd_exp * (1-delta_prob*2)))
    # p['vi'][alt_i] = vi_exp + a_no * (vi_exp - (vi_exp * np.exp(delta_prob)))
    #p['vi'][alt_i] = vi_exp * np.exp(-delta_prob)
    # if delta_prob > 0:
    #p['vd'][alt_i] = vd_exp + (a_go*delta_prob)
    #p['vi'][alt_i] = vi_exp + (a_no*-delta_prob)
    # else:
    p['vd'][alt_i] = vd_exp + (a_go * delta_prob)
    p['vi'][alt_i] = vi_exp + (a_no * -delta_prob)
    #vTrial = vi_exp + (a_go * (vd_exp - vd_exp * np.exp((delta_prob))))
    #vTrial = vi_exp + (a_no * (vi_exp - vi_exp * np.exp((delta_prob))))
    # p['vd'][alt_i] = vd_exp + (vd_exp*a_go)*delta_prob
    # p['vi'][alt_i] = vi_exp + (vi_exp*a_no)*-delta_prob
    return p
